import math

from keras.src.backend import config as backend_config
from keras.src.layers.preprocessing.data_layer import DataLayer
from keras.src.layers.preprocessing.image_preprocessing.bounding_boxes.validation import (  # noqa: E501
    densify_bounding_boxes,
)


class BaseImagePreprocessingLayer(DataLayer):
    _USE_BASE_FACTOR = True
    _FACTOR_BOUNDS = (-1, 1)

    def __init__(
        self, factor=None, bounding_box_format=None, data_format=None, **kwargs
    ):
        super().__init__(**kwargs)
        self.bounding_box_format = bounding_box_format
        self.data_format = backend_config.standardize_data_format(data_format)
        if self._USE_BASE_FACTOR:
            factor = factor or 0.0
            self._set_factor(factor)
        elif factor is not None:
            raise ValueError(
                f"Layer {self.__class__.__name__} does not take "
                f"a `factor` argument. Received: factor={factor}"
            )

    def _set_factor(self, factor):
        error_msg = (
            "The `factor` argument should be a number "
            "(or a list of two numbers) "
            "in the range "
            f"[{self._FACTOR_BOUNDS[0]}, {self._FACTOR_BOUNDS[1]}]. "
            f"Received: factor={factor}"
        )
        if isinstance(factor, (tuple, list)):
            if len(factor) != 2:
                raise ValueError(error_msg)
            if (
                factor[0] > self._FACTOR_BOUNDS[1]
                or factor[1] < self._FACTOR_BOUNDS[0]
            ):
                raise ValueError(error_msg)
            lower, upper = sorted(factor)
        elif isinstance(factor, (int, float)):
            if (
                factor < self._FACTOR_BOUNDS[0]
                or factor > self._FACTOR_BOUNDS[1]
            ):
                raise ValueError(error_msg)
            factor = abs(factor)
            lower, upper = [max(-factor, self._FACTOR_BOUNDS[0]), factor]
        else:
            raise ValueError(error_msg)
        self.factor = lower, upper

    def get_random_transformation(self, data, training=True, seed=None):
        return None

    def _transform_images(self, images, transformation, interpolation):
        raise NotImplementedError()

    def transform_images(self, images, transformation, training=True):
        images = self.backend.cast(images, self.compute_dtype)
        if training:
            images = self._transform_images(
                images, transformation, self.interpolation
            )
        return images

    def transform_labels(self, labels, transformation, training=True):
        raise NotImplementedError()

    def transform_bounding_boxes(
        self,
        bounding_boxes,
        transformation,
        training=True,
    ):
        raise NotImplementedError()

    def transform_segmentation_masks(
        self, segmentation_masks, transformation, training=True
    ):
        if training:
            segmentation_masks = self._transform_images(
                segmentation_masks, transformation, "nearest"
            )
        return segmentation_masks

    def transform_single_image(self, image, transformation, training=True):
        images = self.backend.numpy.expand_dims(image, axis=0)
        outputs = self.transform_images(
            images, transformation=transformation, training=training
        )
        return self.backend.numpy.squeeze(outputs, axis=0)

    def transform_single_label(self, label, transformation, training=True):
        labels = self.backend.numpy.expand_dims(label, axis=0)
        outputs = self.transform_labels(
            labels, transformation=transformation, training=training
        )
        return self.backend.numpy.squeeze(outputs, axis=0)

    def transform_single_bounding_box(
        self,
        bounding_box,
        transformation,
        training=True,
    ):
        bounding_boxes = self._format_single_input_bounding_box(bounding_box)
        outputs = self.transform_bounding_boxes(
            bounding_boxes,
            transformation=transformation,
            training=training,
        )
        bounding_box = self._format_single_output_bounding_box(outputs)
        return bounding_box

    def transform_single_segmentation_mask(
        self, segmentation_mask, transformation, training=True
    ):
        segmentation_masks = self.backend.numpy.expand_dims(
            segmentation_mask, axis=0
        )
        outputs = self.transform_segmentation_masks(
            segmentation_masks, transformation=transformation, training=training
        )
        return self.backend.numpy.squeeze(outputs, axis=0)

    def _is_batched(self, maybe_image_batch):
        shape = self.backend.core.shape(maybe_image_batch)
        if len(shape) == 3:
            return False
        if len(shape) == 4:
            return True
        raise ValueError(
            "Expected image tensor to have rank 3 (single image) "
            f"or 4 (batch of images). Received: data.shape={shape}"
        )

    def call(self, data, training=True):
        transformation = self.get_random_transformation(data, training=training)
        if isinstance(data, dict):
            is_batched = self._is_batched(data["images"])
            if is_batched:
                data["images"] = self.transform_images(
                    self.backend.convert_to_tensor(data["images"]),
                    transformation=transformation,
                    training=training,
                )
            else:
                data["images"] = self.transform_single_image(
                    self.backend.convert_to_tensor(data["images"]),
                    transformation=transformation,
                    training=training,
                )
            if "bounding_boxes" in data:
                if not self.bounding_box_format:
                    raise ValueError(
                        "You passed an input with a 'bounding_boxes' key, "
                        "but you didn't specify a bounding box format. "
                        "Pass a `bounding_box_format` argument to your "
                        f"{self.__class__.__name__} layer, e.g. "
                        "`bounding_box_format='xyxy'`."
                    )
                bounding_boxes = densify_bounding_boxes(
                    data["bounding_boxes"],
                    is_batched=is_batched,
                    backend=self.backend,
                )

                if is_batched:
                    data["bounding_boxes"] = self.transform_bounding_boxes(
                        bounding_boxes,
                        transformation=transformation,
                        training=training,
                    )
                else:
                    data["bounding_boxes"] = self.transform_single_bounding_box(
                        bounding_boxes,
                        transformation=transformation,
                        training=training,
                    )
            if "labels" in data:
                if is_batched:
                    data["labels"] = self.transform_labels(
                        self.backend.convert_to_tensor(data["labels"]),
                        transformation=transformation,
                        training=training,
                    )
                else:
                    data["labels"] = self.transform_single_label(
                        self.backend.convert_to_tensor(data["labels"]),
                        transformation=transformation,
                        training=training,
                    )
            if "segmentation_masks" in data:
                if is_batched:
                    data["segmentation_masks"] = (
                        self.transform_segmentation_masks(
                            data["segmentation_masks"],
                            transformation=transformation,
                            training=training,
                        )
                    )
                else:
                    data["segmentation_masks"] = (
                        self.transform_single_segmentation_mask(
                            data["segmentation_masks"],
                            transformation=transformation,
                            training=training,
                        )
                    )
            return data

        # `data` is just images.
        if self._is_batched(data):
            return self.transform_images(
                self.backend.convert_to_tensor(data),
                transformation=transformation,
                training=training,
            )
        return self.transform_single_image(
            self.backend.convert_to_tensor(data),
            transformation=transformation,
            training=training,
        )

    def _format_single_input_bounding_box(self, bounding_box):
        for key in bounding_box:
            if key == "labels":
                bounding_box[key] = self.backend.numpy.expand_dims(
                    bounding_box[key], axis=0
                )
            if key == "boxes":
                bounding_box[key] = self.backend.numpy.expand_dims(
                    bounding_box[key], axis=0
                )

        return bounding_box

    def _format_single_output_bounding_box(self, bounding_boxes):
        for key in bounding_boxes:
            if key == "labels":
                bounding_boxes[key] = self.backend.numpy.squeeze(
                    bounding_boxes[key], axis=0
                )
            if key == "boxes":
                bounding_boxes[key] = self.backend.numpy.squeeze(
                    bounding_boxes[key], axis=0
                )

        return bounding_boxes

    def get_config(self):
        config = super().get_config()
        if self.bounding_box_format is not None:
            config.update(
                {
                    "bounding_box_format": self.bounding_box_format,
                }
            )
        return config

    def _transform_value_range(
        self, images, original_range, target_range, dtype="float32"
    ):
        """Convert input values from `original_range` to `target_range`.

        This function is intended to be used in preprocessing layers that
        rely upon color values. This allows us to assume internally that
        the input tensor is always in the range `(0, 255)`.

        Args:
            images: the set of images to transform to the target range.
            original_range: the value range to transform from.
            target_range: the value range to transform to.
            dtype: the dtype to compute the conversion with,
                defaults to "float32".

        Returns:
            a new Tensor with values in the target range.

        Example:

        ```python
        original_range = [0, 1]
        target_range = [0, 255]
        images = layer.preprocessing.transform_value_range(
            images,
            original_range,
            target_range
        )
        images = ops.minimum(images + 10, 255)
        images = layer.preprocessing.transform_value_range(
            images,
            target_range,
            original_range
        )
        ```
        """
        if (
            original_range[0] == target_range[0]
            and original_range[1] == target_range[1]
        ):
            return images

        images = self.backend.cast(images, dtype=dtype)
        original_min_value, original_max_value = self._unwrap_value_range(
            original_range, dtype=dtype
        )
        target_min_value, target_max_value = self._unwrap_value_range(
            target_range, dtype=dtype
        )

        # images in the [0, 1] scale
        images = (images - original_min_value) / (
            original_max_value - original_min_value
        )

        scale_factor = target_max_value - target_min_value
        return (images * scale_factor) + target_min_value

    def _unwrap_value_range(self, value_range, dtype="float32"):
        min_value, max_value = value_range
        min_value = self.backend.cast(min_value, dtype=dtype)
        max_value = self.backend.cast(max_value, dtype=dtype)
        return min_value, max_value

    def _compute_affine_matrix(
        self,
        center_x,
        center_y,
        angle,
        translate_x,
        translate_y,
        scale,
        shear_x,
        shear_y,
        height,
        width,
    ):
        """
        #       Scaling          Shear           Rotation
        #     [sx  0   0]    [1   shx  0]   [cos(θ)  -sin(θ)  0]
        # M = [0   sy  0] *  [shy  1   0] * [sin(θ)   cos(θ)  0]
        #     [0   0   1]    [0    0   1]   [0        0       1]

        # a0 = sx * (cos(θ) + shx * sin(θ))
        # a1 = sx * (-sin(θ) + shx * cos(θ))
        # a2 = tx + cx - cx * a0 - cy * a1
        # b0 = sy * (shy * cos(θ) + sin(θ))
        # b1 = sy * (shy * -sin(θ) + cos(θ))
        # b2 = ty + cy - cx * b0 - cy * b1
        """
        ops = self.backend

        degree_to_radian_factor = ops.convert_to_tensor(math.pi / 180.0)

        angle = angle * degree_to_radian_factor
        shear_x = shear_x * degree_to_radian_factor
        shear_y = shear_y * degree_to_radian_factor

        batch_size = ops.shape(angle)[0]
        dtype = angle.dtype
        width = ops.cast(width, dtype)
        height = ops.cast(height, dtype)
        cx = center_x * (width - 1)
        cy = center_y * (height - 1)

        cos_theta = ops.numpy.cos(angle)
        sin_theta = ops.numpy.sin(angle)
        shear_x = ops.numpy.tan(shear_x)
        shear_y = ops.numpy.tan(shear_y)

        a0 = scale * (cos_theta + shear_x * sin_theta)
        a1 = scale * (-sin_theta + shear_x * cos_theta)
        a2 = translate_x + cx - cx * a0 - cy * a1
        b0 = scale * (shear_y * cos_theta + sin_theta)
        b1 = scale * (shear_y * -sin_theta + cos_theta)
        b2 = translate_y + cy - cx * b0 - cy * b1
        affine_matrix = ops.numpy.concatenate(
            [
                a0[:, None],
                a1[:, None],
                a2[:, None],
                b0[:, None],
                b1[:, None],
                b2[:, None],
                ops.numpy.zeros((batch_size, 2)),
            ],
            axis=1,
        )

        return affine_matrix


base_image_preprocessing_transform_example = """
```python
layer = keras.layers.{LayerName}(bounding_box_format="xyxy")
images = np.random.randint(0, 255, (4, 224, 224, 3), dtype="uint8")

bounding_boxes = {
    "boxes": np.array([
        [[10, 20, 100, 150], [50, 60, 200, 250]],
        [[15, 25, 110, 160], [55, 65, 210, 260]],
        [[20, 30, 120, 170], [60, 70, 220, 270]],
        [[25, 35, 130, 180], [65, 75, 230, 280]],
    ], dtype="float32"),
    "labels": np.array([[0, 1], [1, 2], [2, 3], [0, 3]], dtype="int32")
}

labels = keras.ops.one_hot(
    np.array([0, 1, 2, 3]),
    num_classes=4
)

segmentation_masks = np.random.randint(0, 3, (4, 224, 224, 1), dtype="uint8")

output = layer(
    {
        "images": images,
        "bounding_boxes": bounding_boxes,
        "labels": labels,
        "segmentation_masks": segmentation_masks
    },
    training=True
)
```
"""

base_image_preprocessing_color_example = """
```python
layer = keras.layers.{LayerName}(value_range=(0, 255))
images = np.random.randint(0, 255, (8, 224, 224, 3), dtype="uint8")

labels = keras.ops.one_hot(
    np.array([0, 1, 2, 0, 1, 2, 0, 1]),
    num_classes=3
)

segmentation_masks = np.random.randint(0, 3, (8, 224, 224, 1), dtype="uint8")

output = layer(
    {
        "images": images,
        "labels": labels,
        "segmentation_masks": segmentation_masks
    },
    training=True
)
```
"""



import copy
import functools
import logging
import traceback

import numpy as np


@functools.lru_cache()
def patch_tf2onnx():
    """Patches `tf2onnx` to ensure compatibility with numpy>=2.0.0."""

    from onnx import AttributeProto
    from onnx import TensorProto

    from keras.src.utils.module_utils import tf2onnx

    logger = logging.getLogger(tf2onnx.__name__)

    if not hasattr(np, "object"):
        np.object = object

    def patched_rewrite_constant_fold(g, ops):
        """
        We call tensorflow transform with constant folding but in some cases
        tensorflow does fold all constants. Since there are a bunch of ops in
        onnx that use attributes where tensorflow has dynamic inputs, we badly
        want constant folding to work. For cases where tensorflow missed
        something, make another pass over the graph and fix want we care about.
        """
        func_map = {
            "Add": np.add,
            "GreaterEqual": np.greater_equal,
            "Cast": np.asarray,
            "ConcatV2": np.concatenate,
            "Less": np.less,
            "ListDiff": np.setdiff1d,
            "Mul": np.multiply,
            "Pack": np.stack,
            "Range": np.arange,
            "Sqrt": np.sqrt,
            "Sub": np.subtract,
        }
        ops = list(ops)

        keep_looking = True
        while keep_looking:
            keep_looking = False
            for idx, op in enumerate(ops):
                func = func_map.get(op.type)
                if func is None:
                    continue
                if set(op.output) & set(g.outputs):
                    continue
                try:
                    inputs = []
                    for node in op.inputs:
                        if not node.is_const():
                            break
                        inputs.append(node.get_tensor_value(as_list=False))

                    logger.debug(
                        "op name %s, %s, %s",
                        op.name,
                        len(op.input),
                        len(inputs),
                    )
                    if inputs and len(op.input) == len(inputs):
                        logger.info(
                            "folding node type=%s, name=%s" % (op.type, op.name)
                        )
                        if op.type == "Cast":
                            dst = op.get_attr_int("to")
                            np_type = tf2onnx.utils.map_onnx_to_numpy_type(dst)
                            val = np.asarray(*inputs, dtype=np_type)
                        elif op.type == "ConcatV2":
                            axis = inputs[-1]
                            values = inputs[:-1]
                            val = func(tuple(values), axis)
                        elif op.type == "ListDiff":
                            out_type = op.get_attr_int("out_idx")
                            np_type = tf2onnx.utils.map_onnx_to_numpy_type(
                                out_type
                            )
                            val = func(*inputs)
                            val = val.astype(np_type)
                        elif op.type in ["Pack"]:
                            # handle ops that need input array and axis
                            axis = op.get_attr_int("axis")
                            val = func(inputs, axis=axis)
                        elif op.type == "Range":
                            dtype = op.get_attr_int("Tidx")
                            np_type = tf2onnx.utils.map_onnx_to_numpy_type(
                                dtype
                            )
                            val = func(*inputs, dtype=np_type)
                        else:
                            val = func(*inputs)

                        new_node_name = tf2onnx.utils.make_name(op.name)
                        new_output_name = new_node_name
                        old_output_name = op.output[0]
                        old_node_name = op.name
                        logger.debug(
                            "create const node [%s] replacing [%s]",
                            new_node_name,
                            old_node_name,
                        )
                        ops[idx] = g.make_const(new_node_name, val)

                        logger.debug(
                            "replace old output [%s] with new output [%s]",
                            old_output_name,
                            new_output_name,
                        )
                        # need to re-write the consumers input name to use the
                        # const name
                        consumers = g.find_output_consumers(old_output_name)
                        if consumers:
                            for consumer in consumers:
                                g.replace_input(
                                    consumer, old_output_name, new_output_name
                                )

                        # keep looking until there is nothing we can fold.
                        # We keep the graph in topological order so if we
                        # folded, the result might help a following op.
                        keep_looking = True
                except Exception as ex:
                    tb = traceback.format_exc()
                    logger.info("exception: %s, details: %s", ex, tb)
                    # ignore errors

        return ops

    def patched_get_value_attr(self, external_tensor_storage=None):
        """
        Return onnx attr for value property of node.
        Attr is modified to point to external tensor data stored in
        external_tensor_storage, if included.
        """
        a = self._attr["value"]
        if (
            external_tensor_storage is not None
            and self in external_tensor_storage.node_to_modified_value_attr
        ):
            return external_tensor_storage.node_to_modified_value_attr[self]
        if external_tensor_storage is None or a.type != AttributeProto.TENSOR:
            return a

        def prod(x):
            if hasattr(np, "product"):
                return np.product(x)
            else:
                return np.prod(x)

        if (
            prod(a.t.dims)
            > external_tensor_storage.external_tensor_size_threshold
        ):
            a = copy.deepcopy(a)
            tensor_name = (
                f"{self.name.strip()}_{external_tensor_storage.name_counter}"
            )
            for c in '~"#%&*:<>?/\\{|}':
                tensor_name = tensor_name.replace(c, "_")
            external_tensor_storage.name_counter += 1
            external_tensor_storage.name_to_tensor_data[tensor_name] = (
                a.t.raw_data
            )
            external_tensor_storage.node_to_modified_value_attr[self] = a
            a.t.raw_data = b""
            a.t.ClearField("raw_data")
            location = a.t.external_data.add()
            location.key = "location"
            location.value = tensor_name
            a.t.data_location = TensorProto.EXTERNAL
        return a

    tf2onnx.tfonnx.rewrite_constant_fold = patched_rewrite_constant_fold
    tf2onnx.graph.Node.get_value_attr = patched_get_value_attr



"""Utilities for distribution strategy with JAX backend."""

import jax
import numpy as np

from keras.src.random import seed_generator
from keras.src.utils import jax_utils
from keras.src.utils import rng_utils


def list_devices(device_type=None):
    """Return all the available devices based on the device type.

    Note that this should return the global devices in a distributed setting.

    Args:
        device_type: string of `"cpu"`, `"gpu"` or `"tpu"`. Defaults to `"gpu"`
            or `"tpu"` if available when device_type is not provided. Otherwise
            will return the `"cpu"` devices.

    Return:
        List of devices that are available for distribute computation.
    """
    device_type = device_type.lower() if device_type else None
    jax_devices = jax.devices(backend=device_type)
    return [f"{device.platform}:{device.id}" for device in jax_devices]


def get_device_count(device_type=None):
    """Returns the number of available JAX devices.
    Args:
        device_type: Optional device type to count (e.g., "cpu", "gpu", "tpu").
            If `None`, it defaults to counting "gpu" or "tpu" devices if
            available, otherwise it counts "cpu" devices. It does not
            return the sum of all device types.
    Returns:
        int: The total number of JAX devices for the specified type.
    """
    device_type = device_type.lower() if device_type else None
    return jax.device_count(device_type)


def distribute_tensor(tensor, layout):
    """Distribute the tensor based on the layout.

    Note that this function can be used both in eager context, or within a
    jitted function.

    Args:
        tensor: `jax.Array` that need to be distributed.
        layout: `TensorLayout` for the created variable, or a
            JAX-supported layout instance (e.g. `jax.sharding.Sharding`).

    Returns:
        Distributed value.
    """
    # Avoid circular imports.
    from keras.src.distribution import TensorLayout

    if isinstance(layout, TensorLayout):
        layout = layout.backend_layout

    if jax_utils.is_in_jax_tracing_scope(tensor):
        return jax.lax.with_sharding_constraint(tensor, layout)

    # Skip relayout if unnecessary.
    if isinstance(tensor, jax.Array):
        if isinstance(
            layout, jax.sharding.Sharding
        ) and tensor.sharding.is_equivalent_to(layout, ndim=len(tensor.shape)):
            return tensor
        # JAX explicit "layout" support.
        elif hasattr(layout, "layout"):
            current_layout = getattr(tensor, "layout", None)
            if current_layout == layout:
                return tensor
        # JAX explicit "format" support.
        elif hasattr(layout, "format"):
            current_layout = getattr(tensor, "format", None)
            if current_layout == layout:
                return tensor

    return jax.device_put(tensor, layout)


def distribute_data_input(per_process_batch, layout, batch_dim_name):
    """Distribute the input data with the corresponding layout.

    Note that the inputs here is a local worker batch. Within the local worker,
    the data need to be further partitioned to map to each of the devices.

    Args:
        inputs: `jax.Array` that is already sharded to a local process size.
        layout: `TensorLayout` for the distribution information, or a
            `jax.sharding.Sharding` instance.

    Returns:
        A global batch distributed according to `layout`.
    """
    # Avoid circular imports.
    from keras.src.distribution import TensorLayout

    if isinstance(layout, TensorLayout):
        layout = layout.backend_layout

    return jax.make_array_from_process_local_data(layout, per_process_batch)


def initialize_rng():
    """Initializes the global random number generator across processes.

    This is required for consistent initialization in multi-host settings.
    """
    global_seed = rng_utils.get_random_seed()
    # Only set a random seed if not already set
    # via keras.config.set_random_seed()
    if global_seed is None:
        # Generate a random seed on each CPU host and psum them to get a single
        # consistent seed across all processes.
        cpu_devices = jax.devices("cpu")
        num_local_cpu_devices = jax.local_device_count("cpu")
        # Seed must be in range [0, 2^32 - 1], so to ensure proper range and
        # avoid signed integer overflow, we use uint32.
        local_seed = jax.numpy.asarray(
            [seed_generator.make_default_seed()] * num_local_cpu_devices,
            dtype=jax.numpy.uint32,
        )
        # Sum across processes and pull out the first item.
        global_seed = jax.pmap(
            lambda x: jax.lax.psum(x, "all"),
            axis_name="all",
            devices=cpu_devices,
        )(local_seed).item(0)
        # Set the global seed.
        rng_utils.set_random_seed(global_seed)


def initialize(job_addresses, num_processes, process_id):
    if job_addresses and "," in job_addresses:
        # When user provide all the job addresses, we will split and get the
        # first one, which is the coordinator.
        job_addresses = job_addresses.split(",")
        # Do a sanity check to make sure the number of addresses also match
        # the num_processes.
        if num_processes is not None and num_processes != len(job_addresses):
            raise ValueError(
                f"The provided job_addresses {job_addresses} has "
                f"{len(job_addresses)} jobs, but num_processes is "
                f"{num_processes}"
            )
        coordinator_address = job_addresses[0]
    else:
        coordinator_address = job_addresses

    jax.distributed.initialize(
        coordinator_address=coordinator_address,
        num_processes=num_processes,
        process_id=process_id,
    )

    # Ensure the random number generator is initialized across processes.
    initialize_rng()


def num_processes():
    """Return the number of processes for the current distribution setting."""
    return jax.process_count()


def process_id():
    """Return the current process ID for the distribution setting."""
    return jax.process_index()


def _to_backend_device(device_name):
    if isinstance(device_name, jax.Device):
        return device_name
    device_name = str(device_name)
    if ":" not in device_name:
        device_type, device_id = device_name, 0
    else:
        device_type, device_id = device_name.split(":")

    devices = jax.devices(backend=device_type)
    for device in devices:
        if device.platform == device_type and device.id == int(device_id):
            return device
    raise ValueError(f"Device not found: {device_name}")


def _to_backend_mesh(device_mesh):
    """Convert the DeviceMesh to JAX backend specific Mesh.

    Args:
        device_mesh: DeviceMesh instance to convert.

    Returns:
        A `jax.sharding.Mesh` instance.
    """
    shape = device_mesh.devices.shape
    devices = [_to_backend_device(d) for d in device_mesh.devices.flatten()]
    devices = np.array(devices).reshape(shape)
    return jax.sharding.Mesh(devices, device_mesh.axis_names)


def _to_backend_layout(tensor_layout):
    """Convert the TensorLayout to JAX backend specific Sharding.

    Args:
        tensor_layout: TensorLayout instance to convert.

    Returns:
        A `jax.sharding.NamedSharding` instance.
    """
    if tensor_layout.device_mesh is None:
        raise ValueError(
            "Cannot create sharding when device mesh is not set "
            "for TensorLayout."
        )
    partition_spec = jax.sharding.PartitionSpec(*tensor_layout.axes)
    jax_mesh = tensor_layout.device_mesh.backend_mesh
    return jax.sharding.NamedSharding(jax_mesh, partition_spec)



"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""

from keras.src.backend.common.global_state import clear_session as clear_session
from keras.src.backend.common.keras_tensor import (
    is_keras_tensor as is_keras_tensor,
)
from keras.src.backend.common.variables import (
    standardize_dtype as standardize_dtype,
)
from keras.src.layers.preprocessing.feature_space import (
    FeatureSpace as FeatureSpace,
)
from keras.src.ops.operation_utils import get_source_inputs as get_source_inputs
from keras.src.saving.object_registration import (
    CustomObjectScope as CustomObjectScope,
)
from keras.src.saving.object_registration import (
    CustomObjectScope as custom_object_scope,
)
from keras.src.saving.object_registration import (
    get_custom_objects as get_custom_objects,
)
from keras.src.saving.object_registration import (
    get_registered_name as get_registered_name,
)
from keras.src.saving.object_registration import (
    get_registered_object as get_registered_object,
)
from keras.src.saving.object_registration import (
    register_keras_serializable as register_keras_serializable,
)
from keras.src.saving.serialization_lib import (
    deserialize_keras_object as deserialize_keras_object,
)
from keras.src.saving.serialization_lib import (
    serialize_keras_object as serialize_keras_object,
)
from keras.src.trainers.data_adapters.data_adapter_utils import (
    pack_x_y_sample_weight as pack_x_y_sample_weight,
)
from keras.src.trainers.data_adapters.data_adapter_utils import (
    unpack_x_y_sample_weight as unpack_x_y_sample_weight,
)
from keras.src.trainers.data_adapters.py_dataset_adapter import (
    PyDataset as PyDataset,
)
from keras.src.trainers.data_adapters.py_dataset_adapter import (
    PyDataset as Sequence,
)
from keras.src.utils.audio_dataset_utils import (
    audio_dataset_from_directory as audio_dataset_from_directory,
)
from keras.src.utils.config import Config as Config
from keras.src.utils.dataset_utils import split_dataset as split_dataset
from keras.src.utils.file_utils import get_file as get_file
from keras.src.utils.image_dataset_utils import (
    image_dataset_from_directory as image_dataset_from_directory,
)
from keras.src.utils.image_utils import array_to_img as array_to_img
from keras.src.utils.image_utils import img_to_array as img_to_array
from keras.src.utils.image_utils import load_img as load_img
from keras.src.utils.image_utils import save_img as save_img
from keras.src.utils.io_utils import (
    disable_interactive_logging as disable_interactive_logging,
)
from keras.src.utils.io_utils import (
    enable_interactive_logging as enable_interactive_logging,
)
from keras.src.utils.io_utils import (
    is_interactive_logging_enabled as is_interactive_logging_enabled,
)
from keras.src.utils.model_visualization import model_to_dot as model_to_dot
from keras.src.utils.model_visualization import plot_model as plot_model
from keras.src.utils.numerical_utils import normalize as normalize
from keras.src.utils.numerical_utils import to_categorical as to_categorical
from keras.src.utils.progbar import Progbar as Progbar
from keras.src.utils.rng_utils import set_random_seed as set_random_seed
from keras.src.utils.sequence_utils import pad_sequences as pad_sequences
from keras.src.utils.text_dataset_utils import (
    text_dataset_from_directory as text_dataset_from_directory,
)
from keras.src.utils.timeseries_dataset_utils import (
    timeseries_dataset_from_array as timeseries_dataset_from_array,
)
from keras.utils import bounding_boxes as bounding_boxes
from keras.utils import legacy as legacy



import numpy as np
import openvino.opset15 as ov_opset
from openvino import Type

import keras.src.backend.openvino.numpy as onp
from keras.src import backend
from keras.src.backend.common.backend_utils import (
    _get_output_shape_given_tf_padding,
)
from keras.src.backend.openvino.core import OPENVINO_DTYPES
from keras.src.backend.openvino.core import OpenVINOKerasTensor
from keras.src.backend.openvino.core import get_ov_output


def relu(x):
    x = get_ov_output(x)
    return OpenVINOKerasTensor(ov_opset.relu(x).output(0))


def relu6(x):
    x = get_ov_output(x)
    return OpenVINOKerasTensor(ov_opset.clamp(x, 0.0, 6.0).output(0))


def celu(x, alpha=1.0):
    x = get_ov_output(x)
    const_zero = get_ov_output(0.0, x.get_element_type())
    const_alpha = get_ov_output(alpha, x.get_element_type())
    const_one = get_ov_output(1.0, x.get_element_type())
    exp_x_div_alpha = ov_opset.exp(ov_opset.divide(x, const_alpha)).output(0)
    negative_branch = ov_opset.multiply(
        const_alpha, ov_opset.subtract(exp_x_div_alpha, const_one)
    )

    celu_x = ov_opset.add(
        ov_opset.maximum(x, const_zero).output(0),
        ov_opset.minimum(negative_branch, const_zero).output(0),
    )
    return OpenVINOKerasTensor(celu_x.output(0))


def sigmoid(x):
    x = get_ov_output(x)
    return OpenVINOKerasTensor(ov_opset.sigmoid(x).output(0))


def tanh(x):
    x = get_ov_output(x)
    return OpenVINOKerasTensor(ov_opset.tanh(x).output(0))


def tanh_shrink(x):
    x = get_ov_output(x)
    return OpenVINOKerasTensor(ov_opset.subtract(x, ov_opset.tanh(x)).output(0))


def hard_tanh(x):
    x = get_ov_output(x)
    return OpenVINOKerasTensor(ov_opset.clamp(x, -1.0, 1.0).output(0))


def soft_shrink(x, threshold=0.5):
    x = get_ov_output(x)
    et = x.get_element_type()
    thr = get_ov_output(threshold, et)
    zero = get_ov_output(0.0, et)
    abs_x = ov_opset.abs(x)
    sub = ov_opset.subtract(abs_x, thr)
    shrunk = ov_opset.maximum(sub, zero)
    sign = ov_opset.sign(x)
    out = ov_opset.multiply(sign, shrunk)
    return OpenVINOKerasTensor(out.output(0))


def hard_shrink(x, threshold=0.5):
    x = get_ov_output(x)
    et = x.get_element_type()
    thr = get_ov_output(threshold, et)
    zero = get_ov_output(0.0, et)
    cond = ov_opset.greater(ov_opset.abs(x), thr)
    out = ov_opset.select(cond, x, zero)
    return OpenVINOKerasTensor(out.output(0))


def softplus(x):
    x = get_ov_output(x)
    return OpenVINOKerasTensor(ov_opset.softplus(x).output(0))


def softsign(x):
    x = get_ov_output(x)
    return OpenVINOKerasTensor(ov_opset.softsign(x).output(0))


def silu(x):
    x = get_ov_output(x)
    beta = get_ov_output(1.0, x.get_element_type())
    return OpenVINOKerasTensor(ov_opset.swish(x, beta=beta).output(0))


def log_sigmoid(x):
    x = get_ov_output(x)
    neg_x = ov_opset.negative(x)
    return OpenVINOKerasTensor(
        ov_opset.negative(ov_opset.softplus(neg_x)).output(0)
    )


def leaky_relu(x, negative_slope=0.2):
    x = get_ov_output(x)
    slope_const = ov_opset.constant(
        negative_slope, x.get_element_type()
    ).output(0)
    leaky_relu = ov_opset.prelu(x, slope_const).output(0)
    return OpenVINOKerasTensor(leaky_relu)


def sparse_sigmoid(x):
    x = get_ov_output(x)
    et = x.get_element_type()
    one = get_ov_output(1.0, et)
    neg_one = get_ov_output(-1.0, et)
    half = get_ov_output(0.5, et)
    y = ov_opset.minimum(ov_opset.maximum(x, neg_one), one)
    out = ov_opset.multiply(half, ov_opset.add(y, one))
    return OpenVINOKerasTensor(out.output(0))


def hard_sigmoid(x):
    x = get_ov_output(x)
    alpha = get_ov_output(1.0 / 6.0, x.get_element_type())
    beta = get_ov_output(0.5, x.get_element_type())
    return OpenVINOKerasTensor(ov_opset.hard_sigmoid(x, alpha, beta).output(0))


def hard_silu(x):
    hard_sigmoid_output = get_ov_output(hard_sigmoid(x))
    x = get_ov_output(x)
    return OpenVINOKerasTensor(
        ov_opset.multiply(x, hard_sigmoid_output).output(0)
    )


def elu(x, alpha=1.0):
    x = get_ov_output(x)
    return OpenVINOKerasTensor(ov_opset.elu(x, alpha).output(0))


def selu(x):
    alpha = 1.6732632423543772848170429916717
    scale = 1.0507009873554804934193349852946
    x = get_ov_output(x)
    alpha = get_ov_output(alpha, x.get_element_type())
    scale = get_ov_output(scale, x.get_element_type())
    return OpenVINOKerasTensor(ov_opset.selu(x, alpha, scale).output(0))


def gelu(x, approximate=True):
    x = get_ov_output(x)
    approximate_mode = "erf"
    if approximate:
        approximate_mode = "tanh"
    return OpenVINOKerasTensor(ov_opset.gelu(x, approximate_mode).output(0))


def softmax(x, axis=-1):
    x = get_ov_output(x)
    if axis is None:
        x_shape = ov_opset.shape_of(x)
        flatten_shape = ov_opset.constant([-1], Type.i32).output(0)
        flatten_x = ov_opset.reshape(x, flatten_shape, False).output(0)
        softmax_x = ov_opset.softmax(flatten_x, 0).output(0)
        return OpenVINOKerasTensor(
            ov_opset.reshape(softmax_x, x_shape, False).output(0)
        )
    if isinstance(axis, (tuple, list)):
        if not axis:
            return OpenVINOKerasTensor(x)
        axes_const = ov_opset.constant(sorted(axis), Type.i32).output(0)
        x_max = ov_opset.reduce_max(x, axes_const, True).output(0)
        exp_x = ov_opset.exp(ov_opset.subtract(x, x_max).output(0)).output(0)
        sum_exp = ov_opset.reduce_sum(exp_x, axes_const, True).output(0)
        return OpenVINOKerasTensor(ov_opset.divide(exp_x, sum_exp).output(0))
    return OpenVINOKerasTensor(ov_opset.softmax(x, axis).output(0))


def log_softmax(x, axis=-1):
    x = get_ov_output(x)
    if isinstance(axis, (tuple, list)) and not axis:
        return OpenVINOKerasTensor(x)
    restore_shape = None
    if axis is None:
        restore_shape = ov_opset.shape_of(x)
        flatten_shape = ov_opset.constant([-1], Type.i32).output(0)
        x = ov_opset.reshape(x, flatten_shape, False).output(0)
        axes = [0]
    elif isinstance(axis, (tuple, list)):
        axes = sorted(axis)
    else:
        axes = [axis]
    axes_const = ov_opset.constant(axes, Type.i32).output(0)
    x_max = ov_opset.reduce_max(x, axes_const, True).output(0)
    x_shifted = ov_opset.subtract(x, x_max).output(0)
    sum_exp = ov_opset.reduce_sum(
        ov_opset.exp(x_shifted).output(0), axes_const, True
    ).output(0)
    log_sum_exp = ov_opset.log(sum_exp).output(0)
    result = ov_opset.subtract(x_shifted, log_sum_exp).output(0)
    if restore_shape is not None:
        result = ov_opset.reshape(result, restore_shape, False).output(0)
    return OpenVINOKerasTensor(result)


def squareplus(x, b=4):
    x = get_ov_output(x)
    et = x.get_element_type()
    b = get_ov_output(b, et)
    two = get_ov_output(2.0, et)
    x_squared = ov_opset.multiply(x, x)
    inside = ov_opset.add(x_squared, b)
    root = ov_opset.sqrt(inside)
    summed = ov_opset.add(x, root)
    out = ov_opset.divide(summed, two)
    return OpenVINOKerasTensor(out.output(0))


def sparse_plus(x):
    x = get_ov_output(x)
    et = x.get_element_type()
    one = get_ov_output(1.0, et)
    neg_one = get_ov_output(-1.0, et)
    zero = get_ov_output(0.0, et)
    quarter = get_ov_output(0.25, et)
    x_plus_1 = ov_opset.add(x, one)
    quad = ov_opset.multiply(quarter, ov_opset.multiply(x_plus_1, x_plus_1))
    leq_than_neg_one = ov_opset.less_equal(x, neg_one)
    less_than_one = ov_opset.less(x, one)
    out = ov_opset.select(
        leq_than_neg_one,
        zero,
        ov_opset.select(less_than_one, quad, x),
    )
    return OpenVINOKerasTensor(out.output(0))


def threshold(x, threshold, default_value):
    x = get_ov_output(x)
    et = x.get_element_type()
    thr = get_ov_output(threshold, et)
    dv = get_ov_output(default_value, et)
    cond = ov_opset.greater(x, thr)
    out = ov_opset.select(cond, x, dv)
    return OpenVINOKerasTensor(out.output(0))


def max_pool(
    inputs,
    pool_size,
    strides=None,
    padding="valid",
    data_format=None,
):
    num_spatial_dims = (
        get_ov_output(inputs).get_partial_shape().rank.get_length() - 2
    )
    kwargs = {"dilations": [1] * num_spatial_dims}  # required for ov max_pool
    return _pool(
        inputs,
        pool_size,
        ov_opset.max_pool,
        strides,
        padding,
        data_format,
        **kwargs,
    )


def average_pool(
    inputs,
    pool_size,
    strides=None,
    padding="valid",
    data_format=None,
):
    return _pool(
        inputs,
        pool_size,
        ov_opset.avg_pool,
        strides,
        padding,
        data_format,
        exclude_pad=True,
    )


def _compute_adaptive_gather_indices(
    input_dim, output_size, small_window, big_window
):
    """Compute gather indices for the two-pool gather method."""
    window_starts = np.floor(
        np.arange(output_size) * input_dim / output_size
    ).astype(np.int32)
    window_ends = np.minimum(
        np.ceil(np.arange(1, output_size + 1) * input_dim / output_size).astype(
            np.int32
        ),
        input_dim,
    )
    window_starts = np.minimum(window_starts, input_dim - 1)
    window_sizes = window_ends - window_starts
    small_pool_len = max(1, input_dim - small_window + 1)
    return np.where(
        window_sizes == big_window,
        window_starts + small_pool_len,
        window_starts,
    ).tolist()


def _adaptive_pool_ov(
    inputs, output_size, pool_type, data_format, num_spatial_dims
):
    """Shared OpenVINO implementation for adaptive average/max pooling.

    Uses the two-pool gather method: for each spatial axis independently,
    apply pooling with the small and big kernel sizes (stride=1, VALID),
    concatenate the results, then gather the correct output positions.
    """
    if isinstance(output_size, int):
        output_size = (output_size,) * num_spatial_dims

    data_format = backend.standardize_data_format(data_format)
    inputs = get_ov_output(inputs)

    current = _adjust_input(inputs, num_spatial_dims, data_format)

    for spatial_idx in range(num_spatial_dims):
        gather_axis = 2 + spatial_idx
        current_ps = current.get_partial_shape()
        input_dim = current_ps[gather_axis].get_length()
        output_dim = output_size[spatial_idx]

        if input_dim == output_dim:
            continue

        small_w = int(np.ceil(input_dim / output_dim))
        big_w = small_w + 1
        gather_indices = _compute_adaptive_gather_indices(
            input_dim, output_dim, small_w, big_w
        )

        strides = [1] * num_spatial_dims

        small_kernel = [1] * num_spatial_dims
        small_kernel[spatial_idx] = small_w

        if pool_type == "avg":
            small_pool = ov_opset.avg_pool(
                current,
                strides=strides,
                pads_begin=[],
                pads_end=[],
                kernel_shape=small_kernel,
                exclude_pad=True,
                auto_pad="VALID",
            ).output(0)
        else:
            small_pool = ov_opset.max_pool(
                current,
                strides=strides,
                dilations=[1] * num_spatial_dims,
                pads_begin=[],
                pads_end=[],
                kernel_shape=small_kernel,
                auto_pad="VALID",
            ).output(0)

        if big_w <= input_dim:
            big_kernel = [1] * num_spatial_dims
            big_kernel[spatial_idx] = big_w

            if pool_type == "avg":
                big_pool = ov_opset.avg_pool(
                    current,
                    strides=strides,
                    pads_begin=[],
                    pads_end=[],
                    kernel_shape=big_kernel,
                    exclude_pad=True,
                    auto_pad="VALID",
                ).output(0)
            else:
                big_pool = ov_opset.max_pool(
                    current,
                    strides=strides,
                    dilations=[1] * num_spatial_dims,
                    pads_begin=[],
                    pads_end=[],
                    kernel_shape=big_kernel,
                    auto_pad="VALID",
                ).output(0)

            combined = ov_opset.concat(
                [small_pool, big_pool], gather_axis
            ).output(0)
        else:
            # big_w > input_dim: the big pool produces no outputs and all
            # gather indices come from the small pool, so skip the big pool.
            combined = small_pool

        indices_node = ov_opset.constant(gather_indices, Type.i32).output(0)
        axis_node = ov_opset.constant(gather_axis, Type.i32).output(0)
        current = ov_opset.gather(combined, indices_node, axis_node).output(0)

    result = _adjust_outputs(current, num_spatial_dims, data_format)
    return OpenVINOKerasTensor(result)


def adaptive_average_pool(inputs, output_size, data_format=None):
    inputs_ov = get_ov_output(inputs)
    num_spatial_dims = inputs_ov.get_partial_shape().rank.get_length() - 2
    if num_spatial_dims not in (1, 2, 3):
        raise ValueError(
            "adaptive_average_pool supports 1D, 2D, or 3D inputs only."
        )
    return _adaptive_pool_ov(
        inputs, output_size, "avg", data_format, num_spatial_dims
    )


def adaptive_max_pool(inputs, output_size, data_format=None):
    inputs_ov = get_ov_output(inputs)
    num_spatial_dims = inputs_ov.get_partial_shape().rank.get_length() - 2
    if num_spatial_dims not in (1, 2, 3):
        raise ValueError(
            "adaptive_max_pool supports 1D, 2D, or 3D inputs only."
        )
    return _adaptive_pool_ov(
        inputs, output_size, "max", data_format, num_spatial_dims
    )


def _pool(
    inputs,
    pool_size,
    pooling_func,
    strides=None,
    padding="valid",
    data_format=None,
    **kwargs,
):
    data_format = backend.standardize_data_format(data_format)
    inputs = get_ov_output(inputs)

    num_spatial_dims = inputs.get_partial_shape().rank.get_length() - 2
    if isinstance(pool_size, int):
        pool_size = [pool_size] * num_spatial_dims

    if strides is None:
        strides = pool_size

    strides = _adjust_strides_dilation(strides, num_spatial_dims)
    pad_mode, pads_begin, pads_end = _adjust_padding(padding)
    inputs = _adjust_input(inputs, num_spatial_dims, data_format)
    pool_kwargs = {
        "kernel_shape": pool_size,
        "strides": strides,
        "auto_pad": pad_mode,
        "pads_begin": pads_begin,
        "pads_end": pads_end,
        **kwargs,
    }
    pooled = pooling_func(inputs, **pool_kwargs).output(0)
    adjusted_pooled = _adjust_outputs(pooled, num_spatial_dims, data_format)
    return OpenVINOKerasTensor(adjusted_pooled)


def _adjust_strides_dilation(
    x,
    num_spatial_dims,
):
    # Helper function that converts an operand to a spatial operand.
    x = (x,) * num_spatial_dims if isinstance(x, int) else x
    # OpenVINO expects input in NCHW layout
    # x = [1, 1] + list(x)
    x = list(x)
    return x


def _adjust_padding(
    padding,
):
    padding = padding.lower() if isinstance(padding, str) else padding
    if padding == "same":
        return "SAME_UPPER", [], []
    elif padding == "same_lower":
        return "SAME_LOWER", [], []
    elif padding == "valid":
        return "VALID", [], []
    pads_begin = []
    pads_end = []
    for padding_pair in padding:
        pads_begin.append(padding_pair[0])
        pads_end.append(padding_pair[1])
    return "EXPLICIT", pads_begin, pads_end


def _adjust_input(inputs, num_spatial_dims, data_format):
    if data_format == "channels_first":
        return inputs
    if num_spatial_dims == 1:
        permutation = [0, 2, 1]
    elif num_spatial_dims == 2:
        permutation = [0, 3, 1, 2]
    else:
        permutation = [0, 4, 1, 2, 3]
    permutation = ov_opset.constant(permutation, Type.i32)
    return ov_opset.transpose(inputs, permutation).output(0)


def _adjust_kernel(kernel, num_spatial_dims):
    if num_spatial_dims == 1:
        permutation = [2, 1, 0]
    elif num_spatial_dims == 2:
        permutation = [3, 2, 0, 1]
    else:
        permutation = [4, 3, 0, 1, 2]
    permutation = ov_opset.constant(permutation, Type.i32)
    return ov_opset.transpose(kernel, permutation).output(0)


def _adjust_depthwise_kernel(kernel, num_spatial_dims):
    # kernel layout: filter_H, filter_W, C_IN, Ch_mul
    if num_spatial_dims == 1:
        # kernel layout: filter_H, C_IN, Ch_mul
        permutation = [1, 2, 0]
    elif num_spatial_dims == 2:
        # kernel layout: filter_H, filter_W, C_IN, Ch_mul
        permutation = [2, 3, 0, 1]
    else:
        # kernel layout: filter_H, filter_W, filter_Z, C_IN, Ch_mul
        permutation = [3, 4, 0, 1, 2]
    permutation = ov_opset.constant(permutation, Type.i32)
    return ov_opset.transpose(kernel, permutation).output(0)


def _adjust_outputs(outputs, num_spatial_dims, data_format):
    if data_format == "channels_first":
        return outputs
    # convert a tensor from NCHW to NHWC layout
    if num_spatial_dims == 1:
        permutation = [0, 2, 1]
    elif num_spatial_dims == 2:
        permutation = [0, 2, 3, 1]
    else:
        permutation = [0, 2, 3, 4, 1]
    permutation = ov_opset.constant(permutation, Type.i32)
    return ov_opset.transpose(outputs, permutation).output(0)


def conv(
    inputs,
    kernel,
    strides=1,
    padding="valid",
    data_format=None,
    dilation_rate=1,
):
    inputs = get_ov_output(inputs)
    kernel = get_ov_output(kernel)

    data_format = backend.standardize_data_format(data_format)
    num_spatial_dims = inputs.get_partial_shape().rank.get_length() - 2

    if data_format == "channels_last":
        inputs_in_channels = inputs.get_partial_shape()[
            2 + num_spatial_dims - 1
        ]
    else:
        inputs_in_channels = inputs.get_partial_shape()[1]
    kernel_in_channels = kernel.get_partial_shape()[-2]

    strides = _adjust_strides_dilation(strides, num_spatial_dims)
    dilation_rate = _adjust_strides_dilation(dilation_rate, num_spatial_dims)
    pad_mode, pads_begin, pads_end = _adjust_padding(padding)
    inputs = _adjust_input(inputs, num_spatial_dims, data_format)
    kernel = _adjust_kernel(kernel, num_spatial_dims)

    num_groups = (
        inputs_in_channels.get_length() // kernel_in_channels.get_length()
    )
    if num_groups == 1:
        conv = ov_opset.convolution(
            inputs,
            kernel,
            strides,
            pads_begin,
            pads_end,
            dilation_rate,
            pad_mode,
        )
    else:
        input_shape = ov_opset.shape_of(inputs).output(0)
        filter_shape = ov_opset.shape_of(kernel).output(0)
        zero_const = ov_opset.constant([0], Type.i32).output(0)
        one_const = ov_opset.constant([1], Type.i32).output(0)
        two_const = ov_opset.constant([2], Type.i32).output(0)
        input_cin = ov_opset.slice(
            input_shape, one_const, two_const, one_const
        ).output(0)
        filter_cin = ov_opset.slice(
            filter_shape, one_const, two_const, one_const
        ).output(0)
        num_groups = ov_opset.divide(input_cin, filter_cin).output(0)

        # reshape the filter based on the number of groups information
        int_max_const = ov_opset.constant([2**31 - 1], Type.i32).output(0)
        filter_cout = ov_opset.slice(
            filter_shape, zero_const, one_const, one_const
        ).output(0)
        filter_new_cout = ov_opset.divide(filter_cout, num_groups).output(0)
        shape_cin_xy = ov_opset.slice(
            filter_shape, one_const, int_max_const, one_const
        ).output(0)
        filter_new_shape = ov_opset.concat(
            [num_groups, filter_new_cout, shape_cin_xy], 0
        ).output(0)
        new_filter = ov_opset.reshape(kernel, filter_new_shape, False).output(0)
        conv = ov_opset.group_convolution(
            inputs,
            new_filter,
            strides,
            pads_begin,
            pads_end,
            dilation_rate,
            pad_mode,
        )
    conv = _adjust_outputs(conv.output(0), num_spatial_dims, data_format)
    return OpenVINOKerasTensor(conv)


def depthwise_conv(
    inputs,
    kernel,
    strides=1,
    padding="valid",
    data_format=None,
    dilation_rate=1,
):
    inputs = get_ov_output(inputs)
    kernel = get_ov_output(kernel)

    data_format = backend.standardize_data_format(data_format)
    num_spatial_dims = inputs.get_partial_shape().rank.get_length() - 2

    if data_format != "channels_last":
        raise ValueError(
            "OpenVINO depthwise_conv only supports 'channels_last' "
            f"data format. Received: data_format={data_format}"
        )

    strides = _adjust_strides_dilation(strides, num_spatial_dims)
    dilation_rate = _adjust_strides_dilation(dilation_rate, num_spatial_dims)
    pad_mode, pads_begin, pads_end = _adjust_padding(padding)

    inputs = _adjust_input(inputs, num_spatial_dims, data_format)
    kernel = _adjust_depthwise_kernel(kernel, num_spatial_dims)
    unsqueeze_dim = ov_opset.constant([2], Type.i32)
    kernel = ov_opset.unsqueeze(kernel, unsqueeze_dim)

    group_conv = ov_opset.group_convolution(
        inputs, kernel, strides, pads_begin, pads_end, dilation_rate, pad_mode
    )
    group_conv = _adjust_outputs(
        group_conv.output(0), num_spatial_dims, data_format
    )
    return OpenVINOKerasTensor(group_conv)


def separable_conv(
    inputs,
    depthwise_kernel,
    pointwise_kernel,
    strides=1,
    padding="valid",
    data_format=None,
    dilation_rate=1,
):
    data_format = backend.standardize_data_format(data_format)
    depthwise_conv_output = depthwise_conv(
        inputs,
        depthwise_kernel,
        strides,
        padding,
        data_format,
        dilation_rate,
    )
    return conv(
        depthwise_conv_output,
        pointwise_kernel,
        strides=1,
        padding="valid",
        data_format=data_format,
        dilation_rate=dilation_rate,
    )


def conv_transpose(
    inputs,
    kernel,
    strides=1,
    padding="valid",
    output_padding=None,
    data_format=None,
    dilation_rate=1,
):
    inputs = get_ov_output(inputs)
    kernel = get_ov_output(kernel)

    data_format = backend.standardize_data_format(data_format)
    num_spatial_dims = inputs.get_partial_shape().rank.get_length() - 2

    strides = _adjust_strides_dilation(strides, num_spatial_dims)
    dilation_rate = _adjust_strides_dilation(dilation_rate, num_spatial_dims)

    # Convert to channels-first (NCHW) layout
    inputs = _adjust_input(inputs, num_spatial_dims, data_format)
    # Rearrange kernel from Keras (*kernel, C_out, C_in)
    # to OpenVINO format (C_in, C_out, *kernel)
    kernel = _adjust_kernel(kernel, num_spatial_dims)

    # inputs: (N, C_in, *spatial), kernel: (C_in, C_out, *kernel_size)
    input_pshape = inputs.get_partial_shape()
    kernel_pshape = kernel.get_partial_shape()

    spatial_output_shape = []
    all_static = True
    for i in range(num_spatial_dims):
        in_dim = input_pshape[2 + i]
        k_dim = kernel_pshape[2 + i]
        s = strides[i]
        d = dilation_rate[i]
        op_i = (
            output_padding
            if output_padding is None or isinstance(output_padding, int)
            else output_padding[i]
        )
        if in_dim.is_static and k_dim.is_static:
            out_dim = _get_output_shape_given_tf_padding(
                input_size=in_dim.get_length(),
                kernel_size=k_dim.get_length(),
                strides=s,
                padding=padding,
                output_padding=op_i,
                dilation_rate=d,
            )
            spatial_output_shape.append(out_dim)
        else:
            all_static = False
            break

    pad_mode = "SAME_LOWER" if padding.lower() == "same" else "VALID"

    if all_static:
        output_shape_node = ov_opset.constant(
            spatial_output_shape, Type.i64
        ).output(0)
    else:
        output_shape_node = None

    conv_t = ov_opset.convolution_backprop_data(
        inputs,
        kernel,
        strides=strides,
        output_shape=output_shape_node,
        dilations=dilation_rate,
        auto_pad=pad_mode,
    )
    result = _adjust_outputs(conv_t.output(0), num_spatial_dims, data_format)
    return OpenVINOKerasTensor(result)


def one_hot(x, num_classes, axis=-1, dtype=None, sparse=False):
    if sparse:
        raise ValueError("`sparse=True` is not supported with openvino backend")
    x = get_ov_output(x)
    if dtype is None:
        dtype = backend.floatx()
    ov_dtype = OPENVINO_DTYPES[dtype]
    on_value = get_ov_output(1, ov_dtype)
    off_value = get_ov_output(0, ov_dtype)
    one_hot_encoded = ov_opset.one_hot(
        x,
        depth=num_classes,
        axis=axis,
        on_value=on_value,
        off_value=off_value,
    ).output(0)
    return OpenVINOKerasTensor(one_hot_encoded)


def multi_hot(x, num_classes, axis=-1, dtype=None, sparse=False):
    reduction_axis = 1 if len(x.shape) > 1 else 0
    if backend.standardize_dtype(dtype) == "bool":
        outputs = one_hot(x, num_classes, axis=axis, dtype=dtype, sparse=sparse)
        result = ov_opset.reduce_logical_or(outputs, reduction_axis)
    else:
        outputs = one_hot(x, num_classes, axis=axis, dtype=dtype)
        result = ov_opset.reduce_max(outputs, reduction_axis)
    return OpenVINOKerasTensor(result.output(0))


def categorical_crossentropy(target, output, from_logits=False, axis=-1):
    target = get_ov_output(target)
    output = get_ov_output(output)

    if target.shape != output.shape:
        raise ValueError(
            "Arguments `target` and `output` must have the same shape. "
            "Received: "
            f"target.shape={target.shape}, output.shape={output.shape}"
        )
    if len(target.shape) < 1:
        raise ValueError(
            "Arguments `target` and `output` must be at least rank 1. "
            "Received: "
            f"target.shape={target.shape}, output.shape={output.shape}"
        )

    if from_logits:
        log_prob = ov_opset.log_softmax(output, axis).output(0)
    else:
        sum_result = ov_opset.reduce_sum(output, axis, keep_dims=True).output(0)
        output = ov_opset.divide(output, sum_result).output(0)
        output = ov_opset.clamp(
            output, min_value=backend.epsilon(), max_value=1 - backend.epsilon()
        ).output(0)
        log_prob = ov_opset.log(output).output(0)
    result = ov_opset.multiply(target, log_prob).output(0)
    loss = ov_opset.reduce_sum(result, axis).output(0)
    loss = ov_opset.negative(loss).output(0)
    return OpenVINOKerasTensor(loss)


def sparse_categorical_crossentropy(target, output, from_logits=False, axis=-1):
    target = get_ov_output(target)
    # one_hot requires integer indices
    # cast unconditionally, matching JAX/TF/Torch
    target = ov_opset.convert(target, Type.i64).output(0)
    output = get_ov_output(output)

    if len(target.shape) == len(output.shape) and target.shape[-1] == 1:
        target = ov_opset.squeeze(target, -1).output(0)

    if len(output.shape) < 1:
        raise ValueError(
            "Argument `output` must be at least rank 1. "
            "Received: "
            f"output.shape={output.shape}"
        )

    output_shape_without_class_dim = list(output.shape)
    del output_shape_without_class_dim[axis]

    if list(target.shape) != output_shape_without_class_dim:
        raise ValueError(
            "Arguments `target` and `output` must have the same shape "
            "up until the last dimension: "
            f"target.shape={target.shape}, output.shape={output.shape}"
        )

    if from_logits:
        log_prob = ov_opset.log_softmax(output, axis).output(0)
    else:
        sum = ov_opset.reduce_sum(output, axis, keep_dims=True).output(0)
        output = ov_opset.divide(output, sum).output(0)
        output = ov_opset.clamp(
            output, min_value=backend.epsilon(), max_value=1 - backend.epsilon()
        ).output(0)
        log_prob = ov_opset.log(output).output(0)

    output_type = output.get_element_type()
    on_val = ov_opset.constant(1, output_type).output(0)
    off_val = ov_opset.constant(0, output_type).output(0)
    one_hot_target = ov_opset.one_hot(
        target,
        depth=output.shape[axis],
        on_value=on_val,
        off_value=off_val,
        axis=axis,
    ).output(0)
    result = ov_opset.multiply(one_hot_target, log_prob).output(0)
    loss = ov_opset.reduce_sum(result, axis).output(0)
    loss = ov_opset.negative(loss).output(0)
    return OpenVINOKerasTensor(loss)


def binary_crossentropy(target, output, from_logits=False):
    target = get_ov_output(target)
    output = get_ov_output(output)
    if target.get_element_type() != output.get_element_type():
        output = ov_opset.convert(output, target.get_element_type()).output(0)

    if target.shape != output.shape:
        raise ValueError(
            "Arguments `target` and `output` must have the same shape. "
            "Received: "
            f"target.shape={target.shape}, output.shape={output.shape}"
        )

    if from_logits:
        one = ov_opset.constant(1, target.get_element_type()).output(0)
        neg_output = ov_opset.negative(output).output(0)
        one_minus_target = ov_opset.subtract(one, target).output(0)
        bce = ov_opset.add(
            ov_opset.multiply(
                target, ov_opset.softplus(neg_output).output(0)
            ).output(0),
            ov_opset.multiply(
                one_minus_target, ov_opset.softplus(output).output(0)
            ).output(0),
        ).output(0)
        return OpenVINOKerasTensor(bce)

    output = ov_opset.clamp(
        output, min_value=backend.epsilon(), max_value=1 - backend.epsilon()
    ).output(0)
    one = ov_opset.constant(1, target.get_element_type()).output(0)

    minus_output = ov_opset.subtract(one, output).output(0)
    minus_target = ov_opset.subtract(one, target).output(0)

    log_prob = ov_opset.log(output).output(0)
    minus_log_prob = ov_opset.log(minus_output).output(0)
    result = ov_opset.multiply(target, log_prob).output(0)
    minus_result = ov_opset.multiply(minus_target, minus_log_prob).output(0)
    bce = ov_opset.add(result, minus_result).output(0)
    bce = ov_opset.negative(bce).output(0)
    return OpenVINOKerasTensor(bce)


def moments(x, axes, keepdims=False, synchronized=False):
    x = get_ov_output(x)
    ori_type = x.get_element_type()
    if ori_type == Type.f16:
        x = ov_opset.convert(x, Type.f32).output(0)
    axes_c = ov_opset.constant(axes, Type.i32).output(0)
    const_two = ov_opset.constant(2, x.get_element_type()).output(0)
    mean = ov_opset.reduce_mean(x, axes_c, keepdims).output(0)
    squared_x_mean = ov_opset.reduce_mean(
        ov_opset.power(x, const_two).output(0), axes_c, keepdims
    ).output(0)
    variance = ov_opset.subtract(
        squared_x_mean, ov_opset.power(mean, const_two).output(0)
    ).output(0)
    if ori_type == Type.f16:
        fp16_max = float(np.finfo(np.float16).max)
        fp16_min = float(np.finfo(np.float16).min)
        mean = ov_opset.convert(
            ov_opset.clamp(mean, fp16_min, fp16_max).output(0), Type.f16
        ).output(0)
        variance = ov_opset.convert(
            ov_opset.clamp(variance, 0.0, fp16_max).output(0), Type.f16
        ).output(0)
    return OpenVINOKerasTensor(mean), OpenVINOKerasTensor(variance)


def batch_normalization(
    x, mean, variance, axis, offset=None, scale=None, epsilon=1e-3
):
    x = get_ov_output(x)
    mean = get_ov_output(mean)
    variance = get_ov_output(variance)
    if offset is not None:
        offset = get_ov_output(offset)
    else:
        mean_shape = ov_opset.shape_of(mean)
        mean_type = mean.get_element_type()
        zero_const = ov_opset.constant([0], mean_type)
        offset = ov_opset.broadcast(zero_const, mean_shape)
    if scale is not None:
        scale = get_ov_output(scale)
    else:
        mean_shape = ov_opset.shape_of(mean)
        mean_type = mean.get_element_type()
        one_const = ov_opset.constant([1], mean_type)
        scale = ov_opset.broadcast(one_const, mean_shape)

    # adjust x input to have the second dimension representing the channel axis
    x_rank = x.get_partial_shape().rank.get_length()
    if axis < 0:
        axis += x_rank
    if axis != 1:
        perm_vector = list(range(0, x_rank))
        perm_vector[1] = axis
        perm_vector[axis] = 1
        perm_vector = ov_opset.constant(perm_vector, Type.i32).output(0)
        x = ov_opset.transpose(x, perm_vector).output(0)
    batch_norm = ov_opset.batch_norm_inference(
        x, scale, offset, mean, variance, epsilon
    ).output(0)
    if axis != 1:
        perm_vector = list(range(0, x_rank))
        perm_vector[1] = axis
        perm_vector[axis] = 1
        perm_vector = ov_opset.constant(perm_vector, Type.i32).output(0)
        batch_norm = ov_opset.transpose(batch_norm, perm_vector).output(0)
    return OpenVINOKerasTensor(batch_norm)


def ctc_loss(target, output, target_length, output_length, mask_index=0):
    target = get_ov_output(target)
    output = get_ov_output(output)
    target_length = get_ov_output(target_length)
    output_length = get_ov_output(output_length)
    ctc_loss_ = ov_opset.ctc_loss(
        output, output_length, target, target_length, blank_index=mask_index
    )
    ctc_loss_ = ov_opset.convert(ctc_loss_, OPENVINO_DTYPES[backend.floatx()])
    return OpenVINOKerasTensor(ctc_loss_.output(0))


def ctc_decode(
    inputs,
    sequence_lengths,
    strategy="greedy",
    beam_width=100,
    top_paths=1,
    merge_repeated=True,
    mask_index=0,
):
    raise NotImplementedError(
        "`ctc_decode` is not supported with openvino backend"
    )


def psnr(x1, x2, max_val):
    from keras.src.backend.openvino.numpy import log10

    x1 = get_ov_output(x1)
    x2 = get_ov_output(x2)
    max_val = get_ov_output(max_val, x1.get_element_type())
    diff = ov_opset.subtract(x1, x2)
    squared_diff = ov_opset.multiply(diff, diff)
    reduction_axes = list(range(0, x1.get_partial_shape().rank.get_length()))
    mse = ov_opset.reduce_mean(squared_diff, reduction_axes).output(0)
    log_max_val = get_ov_output(log10(OpenVINOKerasTensor(max_val)))
    log_mse = get_ov_output(log10(OpenVINOKerasTensor(mse)))

    psnr = ov_opset.subtract(
        ov_opset.multiply(
            ov_opset.constant(20, log_max_val.get_element_type()), log_max_val
        ),
        ov_opset.multiply(
            ov_opset.constant(10, log_mse.get_element_type()), log_mse
        ),
    ).output(0)
    return OpenVINOKerasTensor(psnr)


def dot_product_attention(
    query,
    key,
    value,
    bias=None,
    mask=None,
    scale=None,
    is_causal=False,
    flash_attention=None,
    attn_logits_soft_cap=None,
):
    if bias is not None:
        raise NotImplementedError(
            "`dot_product_attention` with `bias` is not supported "
            "with openvino backend"
        )
    if flash_attention:
        raise NotImplementedError(
            "`dot_product_attention` with `flash_attention` is not supported "
            "with openvino backend"
        )
    if attn_logits_soft_cap is not None:
        raise NotImplementedError(
            "`dot_product_attention` with `attn_logits_soft_cap` is not "
            "supported with openvino backend"
        )
    query = get_ov_output(query)
    key = get_ov_output(key)
    value = get_ov_output(value)
    if query.get_element_type() != key.get_element_type():
        ov_type = OPENVINO_DTYPES[backend.floatx()]
        query = ov_opset.convert(query, ov_type).output(0)
        key = ov_opset.convert(key, ov_type).output(0)
    if value.get_element_type() != query.get_element_type():
        value = ov_opset.convert(value, query.get_element_type()).output(0)
    axes_const = ov_opset.constant([0, 2, 1, 3], Type.i32).output(0)

    query = ov_opset.transpose(query, axes_const)
    key = ov_opset.transpose(key, axes_const)
    value = ov_opset.transpose(value, axes_const)
    mask = get_ov_output(mask) if mask is not None else None
    scale = (
        get_ov_output(scale, query.get_element_type())
        if scale is not None
        else None
    )
    dpa = ov_opset.scaled_dot_product_attention(
        query, key, value, attention_mask=mask, scale=scale, causal=is_causal
    )
    dpa = ov_opset.transpose(dpa, axes_const)
    return OpenVINOKerasTensor(dpa.output(0))


def unfold(input, kernel_size, dilation=1, padding=0, stride=1):
    def _pair(x):
        return (x, x) if isinstance(x, int) else x

    k = _pair(kernel_size)
    d = _pair(dilation)
    p = _pair(padding)
    s = _pair(stride)

    N, C, H, W = input.shape

    # ---- padding ----
    if any(_ > 0 for _ in p):
        input = onp.pad(input, ((0, 0), (0, 0), (p[0], p[0]), (p[1], p[1])))

    # ---- extract patches ----
    patches = ov_opset.extract_image_patches(
        image=input,
        sizes=[k[0], k[1]],
        strides=[s[0], s[1]],
        rates=[d[0], d[1]],
        auto_pad="VALID",
    )  # (N, kH*kW*C, nH, nW)
    N, D, nH, nW = patches.shape
    patches = ov_opset.reshape(patches, [N, k[0], k[1], C, nH, nW], False)
    patches = ov_opset.transpose(
        patches, [0, 3, 1, 2, 4, 5]
    )  # (N, C, kH, kW, nH, nW)
    patches = ov_opset.reshape(patches, [N, C * k[0] * k[1], nH * nW], False)
    return OpenVINOKerasTensor(patches.output(0))


def fold(x, output_size, kernel_size, dilation=1, padding=0, stride=1):
    def _pair(val):
        return (val, val) if isinstance(val, int) else val

    oH, oW = _pair(output_size)
    kH, kW = _pair(kernel_size)
    dH, dW = _pair(dilation)
    pH, pW = _pair(padding)
    sH, sW = _pair(stride)

    x = get_ov_output(x)

    # Derive CKK and C dynamically from the input shape to support dynamic dims.
    shape_x = ov_opset.shape_of(x).output(0)
    one_i64 = ov_opset.constant([1], Type.i64).output(0)
    two_i64 = ov_opset.constant([2], Type.i64).output(0)
    step_i64 = ov_opset.constant([1], Type.i64).output(0)
    CKK_1d = ov_opset.slice(shape_x, one_i64, two_i64, step_i64).output(0)
    KK_node = ov_opset.constant([kH * kW], Type.i64).output(0)
    C_1d = ov_opset.divide(CKK_1d, KK_node).output(0)
    CKK_scalar = ov_opset.squeeze(CKK_1d, [0]).output(0)

    nH = (oH + 2 * pH - dH * (kH - 1) - 1) // sH + 1
    nW = (oW + 2 * pW - dW * (kW - 1) - 1) // sW + 1

    # Reshape: (N, CKK, L) -> (N, CKK, nH, nW); 0 copies the dim from input.
    new_shape = ov_opset.constant([0, 0, nH, nW], Type.i64).output(0)
    x = ov_opset.reshape(x, new_shape, True).output(0)

    # Build identity kernel dynamically: shape (CKK, CKK) via one_hot,
    # then reshape to (CKK, C, kH, kW).
    indices = ov_opset.range(
        ov_opset.constant(0, Type.i64),
        CKK_scalar,
        ov_opset.constant(1, Type.i64),
        Type.i64,
    ).output(0)
    on_val = ov_opset.constant(1, Type.f32).output(0)
    off_val = ov_opset.constant(0, Type.f32).output(0)
    eye = ov_opset.one_hot(
        indices, depth=CKK_scalar, on_value=on_val, off_value=off_val, axis=-1
    ).output(0)
    kernel_shape = ov_opset.concat(
        [CKK_1d, C_1d, ov_opset.constant([kH, kW], Type.i64).output(0)], axis=0
    ).output(0)
    kernel = ov_opset.reshape(eye, kernel_shape, False).output(0)
    kernel = ov_opset.convert(kernel, x.get_element_type()).output(0)

    oH_pad = oH + 2 * pH
    oW_pad = oW + 2 * pW

    output_shape_node = ov_opset.constant([oH_pad, oW_pad], Type.i64).output(0)
    result = ov_opset.convolution_backprop_data(
        x,
        kernel,
        strides=[sH, sW],
        output_shape=output_shape_node,
        dilations=[dH, dW],
        auto_pad="VALID",
    ).output(0)

    if pH > 0 or pW > 0:
        axes = ov_opset.constant([2, 3], Type.i32).output(0)
        start = ov_opset.constant([pH, pW], Type.i32).output(0)
        stop = ov_opset.constant([oH_pad - pH, oW_pad - pW], Type.i32).output(0)
        step = ov_opset.constant([1, 1], Type.i32).output(0)
        result = ov_opset.slice(result, start, stop, step, axes).output(0)

    return OpenVINOKerasTensor(result)


def depth_to_space(x, block_size, data_format="channels_last"):
    """OpenVINO implementation of depth_to_space (pixel shuffle).

    Rearranges data from depth into blocks of spatial data.

    Args:
        x: 4-D tensor with shape (N, H, W, C) for channels_last or
            (N, C, H, W) for channels_first.
        block_size: An integer specifying the block size.
        data_format: "channels_last" or "channels_first".

    Returns:
        A tensor with shape (N, H*block_size, W*block_size, C/block_size**2)
        for channels_last or (N, C/block_size**2, H*block_size, W*block_size)
        for channels_first.
    """
    x = get_ov_output(x)
    # OpenVINO depth_to_space uses "blocks_first" mode by default
    # and expects NCHW format
    if data_format == "channels_last":
        # Convert NHWC to NCHW
        axes = ov_opset.constant([0, 3, 1, 2], Type.i32).output(0)
        x = ov_opset.transpose(x, axes).output(0)
        result = ov_opset.depth_to_space(x, "blocks_first", block_size).output(
            0
        )
        # Convert back to NHWC
        axes_back = ov_opset.constant([0, 2, 3, 1], Type.i32).output(0)
        result = ov_opset.transpose(result, axes_back).output(0)
    else:
        result = ov_opset.depth_to_space(x, "blocks_first", block_size).output(
            0
        )
    return OpenVINOKerasTensor(result)


def space_to_depth(x, block_size, data_format="channels_last"):
    """OpenVINO implementation of space_to_depth (pixel unshuffle).

    Rearranges blocks of spatial data into depth.

    Args:
        x: 4-D tensor with shape (N, H, W, C) for channels_last or
            (N, C, H, W) for channels_first.
        block_size: An integer specifying the block size.
        data_format: "channels_last" or "channels_first".

    Returns:
        A tensor with shape (N, H/block_size, W/block_size, C*block_size**2)
        for channels_last or (N, C*block_size**2, H/block_size, W/block_size)
        for channels_first.
    """
    x = get_ov_output(x)
    # OpenVINO space_to_depth uses "blocks_first" mode by default
    # and expects NCHW format
    if data_format == "channels_last":
        # Convert NHWC to NCHW
        axes = ov_opset.constant([0, 3, 1, 2], Type.i32).output(0)
        x = ov_opset.transpose(x, axes).output(0)
        result = ov_opset.space_to_depth(x, "blocks_first", block_size).output(
            0
        )
        # Convert back to NHWC
        axes_back = ov_opset.constant([0, 2, 3, 1], Type.i32).output(0)
        result = ov_opset.transpose(result, axes_back).output(0)
    else:
        result = ov_opset.space_to_depth(x, "blocks_first", block_size).output(
            0
        )
    return OpenVINOKerasTensor(result)



import warnings

from keras.src import ops
from keras.src.callbacks.callback import Callback
from keras.src.trainers import compile_utils


class MonitorCallback(Callback):
    """Base class for callbacks that monitor a quantity and evaluates
    improvements.

    This class provides common functionality for callbacks that monitor a
    metric during training to determine whether a condition has been met,
    such as improvement over time. It encapsulates logic for selecting
    the comparison operation based on a `monitor` value and `mode`, and
    computing whether a new value is an improvement.

    It is intended to be subclassed by other callbacks like `ModelCheckpoint`,
    `EarlyStopping`, or `ReduceLROnPlateau`, and is not meant to be used
    directly.

    Arguments:
        monitor: Quantity to be monitored. Defaults to `"val_loss"`.
        mode: One of `{"auto", "min", "max"}`. In `min` mode, training will aim
            to minimize the monitored quantity; in `'max'` mode it will aim to
            maximize it.; in `"auto"` mode, the direction is automatically
            inferred from the name of the monitored quantity. Defaults to
            `"auto"`.
        baseline: Floating point initial "best" value of the metric to be
            monitored. If `None` (default), the first monitored value will be
            used.
        min_delta: Minimum change in the monitored quantity to qualify as an
            improvement, i.e. an absolute change of less than min_delta, will
            count as no improvement. Defaults to `0`.

    Raises:
        ValueError: If `mode='auto'` is selected and the direction of the metric
        cannot be inferred.
    """

    def __init__(
        self,
        monitor="val_loss",
        mode="auto",
        baseline=None,
        min_delta=0,
    ):
        super().__init__()
        if mode not in ["auto", "min", "max"]:
            warnings.warn(
                f"{self.__class__.__name__} mode '{mode}' is unknown, fallback "
                "to auto mode.",
                stacklevel=2,
            )
            mode = "auto"
        self.monitor = monitor
        self.mode = mode
        self.best = baseline
        self.min_delta = abs(min_delta)
        self.monitor_op = None

    def _set_monitor_op(self):
        if self.mode == "min":
            self.monitor_op = ops.less
        elif self.mode == "max":
            self.monitor_op = ops.greater
        else:
            metric_name = self.monitor.removeprefix("val_")
            if metric_name == "loss":
                self.monitor_op = ops.less
            if hasattr(self.model, "metrics"):
                all_metrics = []
                for m in self.model.metrics:
                    if isinstance(
                        m,
                        (
                            compile_utils.CompileMetrics,
                            compile_utils.MetricsList,
                        ),
                    ):
                        all_metrics.extend(m.metrics)
                for m in all_metrics:
                    if m.name == metric_name:
                        if hasattr(m, "_direction"):
                            if m._direction == "up":
                                self.monitor_op = ops.greater
                            else:
                                self.monitor_op = ops.less
            if self.monitor_op is None:
                raise ValueError(
                    f"{self.__class__.__name__} callback received "
                    f"monitor={self.monitor}, but Keras isn't able to "
                    "automatically determine whether that metric should be "
                    "maximized or minimized. Pass `mode='max'` in order to "
                    "monitor based on the highest metric value, or pass "
                    "`mode='min'` in order to use the lowest value."
                )
        if self.monitor_op == ops.less:
            self.min_delta *= -1

    def _is_improvement(self, monitor_value, reference_value):
        if reference_value is None:
            return True
        return self.monitor_op(monitor_value - self.min_delta, reference_value)



"""Orbax checkpoint loading functionality."""

import os

from keras.src import backend
from keras.src.distribution import distribution as get_distribution
from keras.src.utils import file_utils
from keras.src.utils.module_utils import ocp


def is_orbax_checkpoint(filepath):
    """Check if the given path is an Orbax checkpoint directory.

    This function implements custom detection logic instead of relying on
    Orbax APIs which may be unreliable in some environments.
    """
    if not file_utils.exists(filepath) or not file_utils.isdir(filepath):
        return False

    try:
        # List directory contents
        contents = file_utils.listdir(filepath)

        # A set is more efficient for membership testing
        orbax_indicators = {
            "orbax.checkpoint",
            "pytree.orbax-checkpoint",
            "checkpoint_metadata",
        }

        # Fast check for standard files
        if not orbax_indicators.isdisjoint(contents):
            return True

        # Check for step directories or temporary files in a single pass
        return any(
            ".orbax-checkpoint-tmp" in item
            or (
                item.isdigit()
                and file_utils.isdir(file_utils.join(filepath, item))
            )
            for item in contents
        )

    except (OSError, PermissionError):
        # If we can't read the directory, assume it's not a checkpoint
        return False


def find_latest_orbax_checkpoint(checkpoint_dir):
    """Find the latest checkpoint in an Orbax checkpoint directory."""
    checkpointer = ocp.training.Checkpointer(directory=checkpoint_dir)
    latest = checkpointer.latest
    if latest is None:
        raise ValueError(f"No valid checkpoints found in {checkpoint_dir}")
    return os.path.join(checkpoint_dir, str(latest.step))


def build_orbax_abstract_pytree(checkpoint_path, ref_state=None):
    """Build an abstract pytree for Orbax loading with target shardings.

    On JAX with an active distribution, returns a pytree of
    `jax.ShapeDtypeStruct` so that Orbax reshards arrays onto the
    current distribution layout instead of restoring saved shardings.
    On all other backends, or when no distribution is active, returns
    `None` (Orbax will use saved shardings — fine when the topology
    hasn't changed).

    Args:
        checkpoint_path: Path to a specific Orbax checkpoint step
            directory (e.g. `<root>/2`).
        ref_state: Optional reference state tree (from
            `model.get_state_tree()`) whose variables carry the
            target shardings. If `None`, shardings default to
            `None` per leaf (Orbax uses saved shardings).

    Returns:
        A pytree of `jax.ShapeDtypeStruct` matching the checkpoint
        structure, or `None` when resharding is not needed.
    """
    if backend.backend() != "jax":
        return None

    if get_distribution() is None:
        return None

    import jax

    pytree_meta = ocp.pytree_metadata(checkpoint_path).metadata

    def _to_abstract(meta, ref=None):
        """Convert metadata leaf to `jax.ShapeDtypeStruct` with sharding."""
        if hasattr(meta, "shape") and hasattr(meta, "dtype"):
            sharding = getattr(ref, "sharding", None)
            return jax.ShapeDtypeStruct(
                meta.shape, meta.dtype, sharding=sharding
            )
        if isinstance(meta, dict):
            r = ref if isinstance(ref, dict) else {}
            return {k: _to_abstract(v, r.get(k)) for k, v in meta.items()}
        return None

    return {
        key: _to_abstract(val, (ref_state or {}).get(key))
        for key, val in pytree_meta.items()
    }



"""Commonly used math operations not included in NumPy."""

from keras.src import backend
from keras.src.api_export import keras_export
from keras.src.backend import KerasTensor
from keras.src.backend import any_symbolic_tensors
from keras.src.ops.operation import Operation
from keras.src.ops.operation_utils import reduce_shape


def _segment_reduce_validation(data, segment_ids):
    data_shape = data.shape
    segment_ids_shape = segment_ids.shape
    if len(segment_ids_shape) > 1:
        raise ValueError(
            "Argument `segment_ids` should be an 1-D vector, got shape: "
            f"{len(segment_ids_shape)}. Consider either flatten input with "
            "segment_ids.reshape((-1)) and "
            "data.reshape((-1, ) + data.shape[len(segment_ids.shape):]) or "
            "vectorize with vmap."
        )
    if (
        segment_ids_shape[0] is not None
        and data_shape[0] is not None
        and segment_ids_shape[0] != data_shape[0]
    ):
        raise ValueError(
            "Argument `segment_ids` and `data` should have same leading "
            f"dimension. Got {segment_ids_shape} v.s. "
            f"{data_shape}."
        )


class SegmentReduction(Operation):
    def __init__(self, num_segments=None, sorted=False, *, name=None):
        super().__init__(name=name)
        self.num_segments = num_segments
        self.sorted = sorted

    def compute_output_spec(self, data, _):
        output_shape = (self.num_segments,) + tuple(data.shape[1:])
        return KerasTensor(shape=output_shape, dtype=data.dtype)


class SegmentSum(SegmentReduction):
    def call(self, data, segment_ids):
        _segment_reduce_validation(data, segment_ids)
        return backend.math.segment_sum(
            data,
            segment_ids,
            num_segments=self.num_segments,
            sorted=self.sorted,
        )


@keras_export("keras.ops.segment_sum")
def segment_sum(data, segment_ids, num_segments=None, sorted=False):
    """Computes the sum of segments in a tensor.

    Args:
        data: Input tensor.
        segment_ids: A N-D tensor containing segment indices for each
            element in `data`. Num dims for segment ids should be strictly
            smaller or equal to number of dims in data.
        num_segments: An integer representing the total number of
            segments. If not specified, it is inferred from the maximum
            value in `segment_ids`.
        sorted: A boolean indicating whether `segment_ids` is sorted.
            Defaults to `False`.

    Returns:
        A tensor containing the sum of segments, where each element
        represents the sum of the corresponding segment in `data`.

    Example:

    >>> data = keras.ops.convert_to_tensor([1, 2, 10, 20, 100, 200])
    >>> segment_ids = keras.ops.convert_to_tensor([0, 0, 1, 1, 2, 2])
    >>> num_segments = 3
    >>> keras.ops.segment_sum(data, segment_ids,num_segments)
    array([3, 30, 300], dtype=int32)
    """
    _segment_reduce_validation(data, segment_ids)
    if any_symbolic_tensors((data,)):
        return SegmentSum(num_segments, sorted).symbolic_call(data, segment_ids)
    return backend.math.segment_sum(
        data, segment_ids, num_segments=num_segments, sorted=sorted
    )


class SegmentMax(SegmentReduction):
    def call(self, data, segment_ids):
        _segment_reduce_validation(data, segment_ids)
        return backend.math.segment_max(
            data,
            segment_ids,
            num_segments=self.num_segments,
            sorted=self.sorted,
        )


@keras_export("keras.ops.segment_max")
def segment_max(data, segment_ids, num_segments=None, sorted=False):
    """Computes the max of segments in a tensor.

    Args:
        data: Input tensor.
        segment_ids: A N-D tensor containing segment indices for each
            element in `data`. data.shape[:len(segment_ids.shape)] should match.
        num_segments: An integer representing the total number of
            segments. If not specified, it is inferred from the maximum
            value in `segment_ids`.
        sorted: A boolean indicating whether `segment_ids` is sorted.
            Defaults to `False`.

    Returns:
        A tensor containing the max of segments, where each element
        represents the max of the corresponding segment in `data`.

    Example:

    >>> data = keras.ops.convert_to_tensor([1, 2, 10, 20, 100, 200])
    >>> segment_ids = keras.ops.convert_to_tensor([0, 0, 1, 1, 2, 2])
    >>> num_segments = 3
    >>> keras.ops.segment_max(data, segment_ids, num_segments)
    array([2, 20, 200], dtype=int32)
    """
    _segment_reduce_validation(data, segment_ids)
    if any_symbolic_tensors((data,)):
        return SegmentMax(num_segments, sorted).symbolic_call(data, segment_ids)
    return backend.math.segment_max(
        data, segment_ids, num_segments=num_segments, sorted=sorted
    )


class TopK(Operation):
    def __init__(self, k, sorted=True, *, name=None):
        super().__init__(name=name)
        self.k = k
        self.sorted = sorted

    def compute_output_spec(self, x):
        output_shape = list(x.shape)
        output_shape[-1] = self.k
        # Return a tuple (values, indices).
        return (
            KerasTensor(shape=output_shape, dtype=x.dtype),
            KerasTensor(shape=output_shape, dtype="int32"),
        )

    def call(self, x):
        return backend.math.top_k(x, self.k, self.sorted)


@keras_export("keras.ops.top_k")
def top_k(x, k, sorted=True):
    """Finds the top-k values and their indices in a tensor.

    Args:
        x: Input tensor.
        k: An integer representing the number of top elements to retrieve.
        sorted: A boolean indicating whether to sort the output in
        descending order. Defaults to `True`.

    Returns:
        A tuple containing two tensors. The first tensor contains the
        top-k values, and the second tensor contains the indices of the
        top-k values in the input tensor.

    Example:

    >>> x = keras.ops.convert_to_tensor([5, 2, 7, 1, 9, 3])
    >>> values, indices = top_k(x, k=3)
    >>> print(values)
    array([9 7 5], shape=(3,), dtype=int32)
    >>> print(indices)
    array([4 2 0], shape=(3,), dtype=int32)

    """
    if any_symbolic_tensors((x,)):
        return TopK(k, sorted).symbolic_call(x)
    return backend.math.top_k(x, k, sorted)


class InTopK(Operation):
    def __init__(self, k, *, name=None):
        super().__init__(name=name)
        self.k = k

    def compute_output_spec(self, targets, predictions):
        return KerasTensor(shape=targets.shape, dtype="bool")

    def call(self, targets, predictions):
        return backend.math.in_top_k(targets, predictions, self.k)


@keras_export("keras.ops.in_top_k")
def in_top_k(targets, predictions, k):
    """Checks if the targets are in the top-k predictions.

    Args:
        targets: A tensor of true labels.
        predictions: A tensor of predicted labels.
        k: An integer representing the number of predictions to consider.

    Returns:
        A boolean tensor of the same shape as `targets`, where each element
        indicates whether the corresponding target is in the top-k predictions.

    Example:

    >>> targets = keras.ops.convert_to_tensor([2, 5, 3])
    >>> predictions = keras.ops.convert_to_tensor(
    ... [[0.1, 0.4, 0.6, 0.9, 0.5],
    ...  [0.1, 0.7, 0.9, 0.8, 0.3],
    ...  [0.1, 0.6, 0.9, 0.9, 0.5]])
    >>> in_top_k(targets, predictions, k=3)
    array([ True False  True], shape=(3,), dtype=bool)
    """
    if any_symbolic_tensors((targets, predictions)):
        return InTopK(k).symbolic_call(targets, predictions)
    return backend.math.in_top_k(targets, predictions, k)


class Logsumexp(Operation):
    def __init__(self, axis=None, keepdims=False, *, name=None):
        super().__init__(name=name)
        self.axis = axis
        self.keepdims = keepdims

    def compute_output_spec(self, x):
        output_shape = reduce_shape(x.shape, self.axis, self.keepdims)
        return KerasTensor(shape=output_shape)

    def call(self, x):
        return backend.math.logsumexp(x, axis=self.axis, keepdims=self.keepdims)


@keras_export("keras.ops.logsumexp")
def logsumexp(x, axis=None, keepdims=False):
    """Computes the logarithm of sum of exponentials of elements in a tensor.

    Args:
        x: Input tensor.
        axis: An integer or a tuple of integers specifying the axis/axes
            along which to compute the sum. If `None`, the sum is computed
            over all elements. Defaults to `None`.
        keepdims: A boolean indicating whether to keep the dimensions of
            the input tensor when computing the sum. Defaults to `False`.

    Returns:
        A tensor containing the logarithm of the sum of exponentials of
        elements in `x`.

    Example:

    >>> x = keras.ops.convert_to_tensor([1., 2., 3.])
    >>> logsumexp(x)
    3.407606
    """
    if any_symbolic_tensors((x,)):
        return Logsumexp(axis, keepdims).symbolic_call(x)
    return backend.math.logsumexp(x, axis=axis, keepdims=keepdims)


class ExtractSequences(Operation):
    def __init__(self, sequence_length, sequence_stride, *, name=None):
        super().__init__(name=name)
        self.sequence_length = sequence_length
        self.sequence_stride = sequence_stride

    def compute_output_spec(self, x):
        if len(x.shape) < 1:
            raise ValueError(
                f"Input should have rank >= 1. "
                f"Received: input.shape = {x.shape}"
            )
        if x.shape[-1] is not None:
            num_sequences = (
                1 + (x.shape[-1] - self.sequence_length) // self.sequence_stride
            )
        else:
            num_sequences = None
        new_shape = x.shape[:-1] + (num_sequences, self.sequence_length)
        return KerasTensor(shape=new_shape, dtype=x.dtype)

    def call(self, x):
        return backend.math.extract_sequences(
            x,
            sequence_length=self.sequence_length,
            sequence_stride=self.sequence_stride,
        )


@keras_export("keras.ops.extract_sequences")
def extract_sequences(x, sequence_length, sequence_stride):
    """Expands the dimension of last axis into sequences of `sequence_length`.

    Slides a window of size `sequence_length` over the last axis of the input
    with a stride of `sequence_stride`, replacing the last axis with
    `[num_sequences, sequence_length]` sequences.

    If the dimension along the last axis is N, the number of sequences can be
    computed by:

    `num_sequences = 1 + (N - sequence_length) // sequence_stride`

    Args:
        x: Input tensor.
        sequence_length: An integer representing the sequences length.
        sequence_stride: An integer representing the sequences hop size.

    Returns:
        A tensor of sequences with shape [..., num_sequences, sequence_length].

    Example:

    >>> x = keras.ops.convert_to_tensor([1, 2, 3, 4, 5, 6])
    >>> extract_sequences(x, 3, 2)
    array([[1, 2, 3],
       [3, 4, 5]])
    """
    if any_symbolic_tensors((x,)):
        return ExtractSequences(sequence_length, sequence_stride).symbolic_call(
            x
        )
    return backend.math.extract_sequences(x, sequence_length, sequence_stride)


class FFT(Operation):
    def compute_output_spec(self, x):
        if not isinstance(x, (tuple, list)) or len(x) != 2:
            raise ValueError(
                "Input `x` should be a tuple of two tensors - real and "
                f"imaginary. Received: x={x}"
            )

        real, imag = x
        # Both real and imaginary parts should have the same shape.
        if real.shape != imag.shape:
            raise ValueError(
                "Input `x` should be a tuple of two tensors - real and "
                "imaginary. Both the real and imaginary parts should have the "
                f"same shape. Received: x[0].shape = {real.shape}, "
                f"x[1].shape = {imag.shape}"
            )

        # We are calculating 1D FFT. Hence, rank >= 1.
        if len(real.shape) < 1:
            raise ValueError(
                f"Input should have rank >= 1. "
                f"Received: input.shape = {real.shape}"
            )

        # The axis along which we are calculating FFT should be fully-defined.
        m = real.shape[-1]
        if m is None:
            raise ValueError(
                f"Input should have its last dimension fully-defined. "
                f"Received: input.shape = {real.shape}"
            )

        return (
            KerasTensor(shape=real.shape, dtype=real.dtype),
            KerasTensor(shape=imag.shape, dtype=imag.dtype),
        )

    def call(self, x):
        return backend.math.fft(x)


@keras_export("keras.ops.fft")
def fft(x):
    """Computes the Fast Fourier Transform along last axis of input.

    Args:
        x: Tuple of the real and imaginary parts of the input tensor. Both
            tensors in the tuple should be of floating type.

    Returns:
        A tuple containing two tensors - the real and imaginary parts of the
        output tensor.

    Example:

    >>> x = (
    ...     keras.ops.convert_to_tensor([1., 2.]),
    ...     keras.ops.convert_to_tensor([0., 1.]),
    ... )
    >>> fft(x)
    (array([ 3., -1.], dtype=float32), array([ 1., -1.], dtype=float32))
    """
    if any_symbolic_tensors(x):
        return FFT().symbolic_call(x)
    return backend.math.fft(x)


class FFT2(Operation):
    def compute_output_spec(self, x):
        axes = (-2, -1)
        if not isinstance(x, (tuple, list)) or len(x) != 2:
            raise ValueError(
                "Input `x` should be a tuple of two tensors - real and "
                f"imaginary. Received: x={x}"
            )

        real, imag = x
        # Both real and imaginary parts should have the same shape.
        if real.shape != imag.shape:
            raise ValueError(
                "Input `x` should be a tuple of two tensors - real and "
                "imaginary. Both the real and imaginary parts should have the "
                f"same shape. Received: x[0].shape = {real.shape}, "
                f"x[1].shape = {imag.shape}"
            )
        # We are calculating 2D FFT. Hence, rank >= 2.
        if len(real.shape) < 2:
            raise ValueError(
                f"Input should have rank >= 2. "
                f"Received: input.shape = {real.shape}"
            )

        # The axes along which we are calculating FFT should be fully-defined.
        m = real.shape[axes[0]]
        n = real.shape[axes[1]]
        if m is None or n is None:
            raise ValueError(
                f"Input should have its {axes} axes fully-defined. "
                f"Received: input.shape = {real.shape}"
            )

        return (
            KerasTensor(shape=real.shape, dtype=real.dtype),
            KerasTensor(shape=imag.shape, dtype=imag.dtype),
        )

    def call(self, x):
        return backend.math.fft2(x)


@keras_export("keras.ops.fft2")
def fft2(x):
    """Computes the 2D Fast Fourier Transform along the last two axes of input.

    Args:
        x: Tuple of the real and imaginary parts of the input tensor. Both
            tensors in the tuple should be of floating type.

    Returns:
        A tuple containing two tensors - the real and imaginary parts of the
        output.

    Example:

    >>> x = (
    ...     keras.ops.convert_to_tensor([[1., 2.], [2., 1.]]),
    ...     keras.ops.convert_to_tensor([[0., 1.], [1., 0.]]),
    ... )
    >>> fft2(x)
    (array([[ 6.,  0.],
        [ 0., -2.]], dtype=float32), array([[ 2.,  0.],
        [ 0., -2.]], dtype=float32))
    """
    if any_symbolic_tensors(x):
        return FFT2().symbolic_call(x)
    return backend.math.fft2(x)


class IFFT2(Operation):
    def compute_output_spec(self, x):
        axes = (-2, -1)
        if not isinstance(x, (tuple, list)) or len(x) != 2:
            raise ValueError(
                "Input `x` should be a tuple of two tensors - real and "
                f"imaginary. Received: x={x}"
            )

        real, imag = x
        # Both real and imaginary parts should have the same shape.
        if real.shape != imag.shape:
            raise ValueError(
                "Input `x` should be a tuple of two tensors - real and "
                "imaginary. Both the real and imaginary parts should have the "
                f"same shape. Received: x[0].shape = {real.shape}, "
                f"x[1].shape = {imag.shape}"
            )
        # We are calculating 2D IFFT. Hence, rank >= 2.
        if len(real.shape) < 2:
            raise ValueError(
                f"Input should have rank >= 2. "
                f"Received: input.shape = {real.shape}"
            )

        # The axes along which we are calculating IFFT should be fully-defined.
        m = real.shape[axes[0]]
        n = real.shape[axes[1]]
        if m is None or n is None:
            raise ValueError(
                f"Input should have its {axes} axes fully-defined. "
                f"Received: input.shape = {real.shape}"
            )

        return (
            KerasTensor(shape=real.shape, dtype=real.dtype),
            KerasTensor(shape=imag.shape, dtype=imag.dtype),
        )

    def call(self, x):
        return backend.math.ifft2(x)


@keras_export("keras.ops.ifft2")
def ifft2(x):
    """Computes the 2D Inverse Fast Fourier Transform along the last two axes of
        input.

    Args:
        x: Tuple of the real and imaginary parts of the input tensor. Both
            tensors in the tuple should be of floating type.

    Returns:
        A tuple containing two tensors - the real and imaginary parts of the
        output.

    Example:

    >>> x = (
    ...     keras.ops.convert_to_tensor([[1., 2.], [2., 1.]]),
    ...     keras.ops.convert_to_tensor([[0., 1.], [1., 0.]]),
    ... )
    >>> ifft2(x)
    (array([[ 6.,  0.],
        [ 0., -2.]], dtype=float32), array([[ 2.,  0.],
        [ 0., -2.]], dtype=float32))
    """
    if any_symbolic_tensors(x):
        return IFFT2().symbolic_call(x)
    return backend.math.ifft2(x)


class RFFT(Operation):
    def __init__(self, fft_length=None, *, name=None):
        super().__init__(name=name)
        self.fft_length = fft_length

    def compute_output_spec(self, x):
        # We are calculating 1D RFFT. Hence, rank >= 1.
        if len(x.shape) < 1:
            raise ValueError(
                f"Input should have rank >= 1. "
                f"Received: input.shape = {x.shape}"
            )

        if self.fft_length is not None:
            new_last_dimension = self.fft_length // 2 + 1
        else:
            if x.shape[-1] is not None:
                new_last_dimension = x.shape[-1] // 2 + 1
            else:
                new_last_dimension = None
        new_shape = x.shape[:-1] + (new_last_dimension,)

        return (
            KerasTensor(shape=new_shape, dtype=x.dtype),
            KerasTensor(shape=new_shape, dtype=x.dtype),
        )

    def call(self, x):
        return backend.math.rfft(x, fft_length=self.fft_length)


@keras_export("keras.ops.rfft")
def rfft(x, fft_length=None):
    """Real-valued Fast Fourier Transform along the last axis of the input.

    Computes the 1D Discrete Fourier Transform of a real-valued signal over the
    inner-most dimension of input.

    Since the Discrete Fourier Transform of a real-valued signal is
    Hermitian-symmetric, RFFT only returns the `fft_length / 2 + 1` unique
    components of the FFT: the zero-frequency term, followed by the
    `fft_length / 2` positive-frequency terms.

    Along the axis RFFT is computed on, if `fft_length` is smaller than the
    corresponding dimension of the input, the dimension is cropped. If it is
    larger, the dimension is padded with zeros.

    Args:
        x: Input tensor.
        fft_length: An integer representing the number of the fft length. If not
            specified, it is inferred from the length of the last axis of `x`.
            Defaults to `None`.

    Returns:
        A tuple containing two tensors - the real and imaginary parts of the
        output.

    Examples:

    >>> x = keras.ops.convert_to_tensor([0.0, 1.0, 2.0, 3.0, 4.0])
    >>> rfft(x)
    (array([10.0, -2.5, -2.5]), array([0.0, 3.4409548, 0.81229924]))

    >>> rfft(x, 3)
    (array([3.0, -1.5]), array([0.0, 0.8660254]))
    """
    if any_symbolic_tensors((x,)):
        return RFFT(fft_length).symbolic_call(x)
    return backend.math.rfft(x, fft_length)


class IRFFT(Operation):
    def __init__(self, fft_length=None, *, name=None):
        super().__init__(name=name)
        self.fft_length = fft_length

    def compute_output_spec(self, x):
        if not isinstance(x, (tuple, list)) or len(x) != 2:
            raise ValueError(
                "Input `x` should be a tuple of two tensors - real and "
                f"imaginary. Received: x={x}"
            )
        real, imag = x
        # Both real and imaginary parts should have the same shape.
        if real.shape != imag.shape:
            raise ValueError(
                "Input `x` should be a tuple of two tensors - real and "
                "imaginary. Both the real and imaginary parts should have the "
                f"same shape. Received: x[0].shape = {real.shape}, "
                f"x[1].shape = {imag.shape}"
            )
        # We are calculating 1D IRFFT. Hence, rank >= 1.
        if len(real.shape) < 1:
            raise ValueError(
                f"Input should have rank >= 1. "
                f"Received: input.shape = {real.shape}"
            )

        if self.fft_length is not None:
            new_last_dimension = self.fft_length
        else:
            if real.shape[-1] is not None:
                new_last_dimension = 2 * (real.shape[-1] - 1)
            else:
                new_last_dimension = None
        new_shape = real.shape[:-1] + (new_last_dimension,)
        return KerasTensor(shape=new_shape, dtype=real.dtype)

    def call(self, x):
        return backend.math.irfft(x, fft_length=self.fft_length)


@keras_export("keras.ops.irfft")
def irfft(x, fft_length=None):
    """Inverse real-valued Fast Fourier transform along the last axis.

    Computes the inverse 1D Discrete Fourier Transform of a real-valued signal
    over the inner-most dimension of input.

    The inner-most dimension of the input is assumed to be the result of RFFT:
    the `fft_length / 2 + 1` unique components of the DFT of a real-valued
    signal. If `fft_length` is not provided, it is computed from the size of the
    inner-most dimension of the input `(fft_length = 2 * (inner - 1))`. If the
    FFT length used to compute is odd, it should be provided since it cannot
    be inferred properly.

    Along the axis IRFFT is computed on, if `fft_length / 2 + 1` is smaller than
    the corresponding dimension of the input, the dimension is cropped. If it is
    larger, the dimension is padded with zeros.

    Args:
        x: Tuple of the real and imaginary parts of the input tensor. Both
            tensors in the tuple should be of floating type.
        fft_length: An integer representing the number of the fft length. If not
            specified, it is inferred from the length of the last axis of `x`.
            Defaults to `None`.

    Returns:
        A tensor containing the inverse real-valued Fast Fourier Transform
        along the last axis of `x`.

    Examples:

    >>> real = keras.ops.convert_to_tensor([0.0, 1.0, 2.0, 3.0, 4.0])
    >>> imag = keras.ops.convert_to_tensor([0.0, 1.0, 2.0, 3.0, 4.0])
    >>> irfft((real, imag))
    array([0.66666667, -0.9106836, 0.24401694])

    >>> irfft(rfft(real, 5), 5)
    array([0.0, 1.0, 2.0, 3.0, 4.0])
    """
    if any_symbolic_tensors(x):
        return IRFFT(fft_length).symbolic_call(x)
    return backend.math.irfft(x, fft_length)


class STFT(Operation):
    def __init__(
        self,
        sequence_length,
        sequence_stride,
        fft_length,
        window="hann",
        center=True,
        *,
        name=None,
    ):
        super().__init__(name=name)
        self.sequence_length = sequence_length
        self.sequence_stride = sequence_stride
        self.fft_length = fft_length
        self.window = window
        self.center = center

    def compute_output_spec(self, x):
        if x.shape[-1] is not None:
            padded = 0 if self.center is False else (self.fft_length // 2) * 2
            num_sequences = (
                1
                + (x.shape[-1] + padded - self.fft_length)
                // self.sequence_stride
            )
        else:
            num_sequences = None
        new_shape = x.shape[:-1] + (num_sequences, self.fft_length // 2 + 1)
        return (
            KerasTensor(shape=new_shape, dtype=x.dtype),
            KerasTensor(shape=new_shape, dtype=x.dtype),
        )

    def call(self, x):
        return backend.math.stft(
            x,
            sequence_length=self.sequence_length,
            sequence_stride=self.sequence_stride,
            fft_length=self.fft_length,
            window=self.window,
            center=self.center,
        )


@keras_export("keras.ops.stft")
def stft(
    x, sequence_length, sequence_stride, fft_length, window="hann", center=True
):
    """Short-Time Fourier Transform along the last axis of the input.

    The STFT computes the Fourier transform of short overlapping windows of the
    input. This giving frequency components of the signal as they change over
    time.

    Args:
        x: Input tensor.
        sequence_length: An integer representing the sequence length.
        sequence_stride: An integer representing the sequence hop size.
        fft_length: An integer representing the size of the FFT to apply. If not
            specified, uses the smallest power of 2 enclosing `sequence_length`.
        window: A string, a tensor of the window or `None`. If `window` is a
            string, available values are `"hann"` and `"hamming"`. If `window`
            is a tensor, it will be used directly as the window and its length
            must be `sequence_length`. If `window` is `None`, no windowing is
            used. Defaults to `"hann"`.
        center: Whether to pad `x` on both sides so that the t-th sequence is
            centered at time `t * sequence_stride`. Otherwise, the t-th sequence
            begins at time `t * sequence_stride`. Defaults to `True`.

    Returns:
        A tuple containing two tensors - the real and imaginary parts of the
        STFT output.

    Example:

    >>> x = keras.ops.convert_to_tensor([0.0, 1.0, 2.0, 3.0, 4.0])
    >>> stft(x, 3, 2, 3)
    (array([[0.75, -0.375],
       [3.75, -1.875],
       [5.25, -2.625]]), array([[0.0, 0.64951905],
       [0.0, 0.64951905],
       [0.0, -0.64951905]]))
    """
    if not isinstance(sequence_stride, int) or sequence_stride <= 0:
        raise ValueError(
            "`sequence_stride` must be a positive integer. "
            f"Received: sequence_stride={sequence_stride}"
        )
    if any_symbolic_tensors((x,)):
        return STFT(
            sequence_length=sequence_length,
            sequence_stride=sequence_stride,
            fft_length=fft_length,
            window=window,
            center=center,
        ).symbolic_call(x)
    return backend.math.stft(
        x,
        sequence_length=sequence_length,
        sequence_stride=sequence_stride,
        fft_length=fft_length,
        window=window,
        center=center,
    )


class ISTFT(Operation):
    def __init__(
        self,
        sequence_length,
        sequence_stride,
        fft_length,
        length=None,
        window="hann",
        center=True,
        *,
        name=None,
    ):
        super().__init__(name=name)
        self.sequence_length = sequence_length
        self.sequence_stride = sequence_stride
        self.fft_length = fft_length
        self.length = length
        self.window = window
        self.center = center

    def compute_output_spec(self, x):
        if not isinstance(x, (tuple, list)) or len(x) != 2:
            raise ValueError(
                "Input `x` should be a tuple of two tensors - real and "
                f"imaginary. Received: x={x}"
            )
        real, imag = x
        # Both real and imaginary parts should have the same shape.
        if real.shape != imag.shape:
            raise ValueError(
                "Input `x` should be a tuple of two tensors - real and "
                "imaginary. Both the real and imaginary parts should have the "
                f"same shape. Received: x[0].shape = {real.shape}, "
                f"x[1].shape = {imag.shape}"
            )
        if len(real.shape) < 2:
            raise ValueError(
                f"Input should have rank >= 2. "
                f"Received: input.shape = {real.shape}"
            )
        if real.shape[-2] is not None:
            output_size = (
                real.shape[-2] - 1
            ) * self.sequence_stride + self.fft_length
            if self.length is not None:
                output_size = self.length
            elif self.center:
                output_size = output_size - (self.fft_length // 2) * 2
        else:
            output_size = None
            if self.length is not None:
                output_size = self.length
        new_shape = real.shape[:-2] + (output_size,)
        return KerasTensor(shape=new_shape, dtype=real.dtype)

    def call(self, x):
        return backend.math.istft(
            x,
            sequence_length=self.sequence_length,
            sequence_stride=self.sequence_stride,
            fft_length=self.fft_length,
            length=self.length,
            window=self.window,
            center=self.center,
        )


@keras_export("keras.ops.istft")
def istft(
    x,
    sequence_length,
    sequence_stride,
    fft_length,
    length=None,
    window="hann",
    center=True,
):
    """Inverse Short-Time Fourier Transform along the last axis of the input.

    To reconstruct an original waveform, the parameters should be the same in
    `stft`.

    Args:
        x: Tuple of the real and imaginary parts of the input tensor. Both
            tensors in the tuple should be of floating type.
        sequence_length: An integer representing the sequence length.
        sequence_stride: An integer representing the sequence hop size.
        fft_length: An integer representing the size of the FFT that produced
            `stft`. Should be of type `int32`.
        length: An integer representing the output is clipped to exactly length.
            If not specified, no padding or clipping take place. Defaults to
            `None`.
        window: A string, a tensor of the window or `None`. If `window` is a
            string, available values are `"hann"` and `"hamming"`. If `window`
            is a tensor, it will be used directly as the window and its length
            must be `sequence_length`. If `window` is `None`, no windowing is
            used. Defaults to `"hann"`.
        center: Whether `x` was padded on both sides so that the t-th sequence
            is centered at time `t * sequence_stride`. Defaults to `True`.

    Returns:
        A tensor containing the inverse Short-Time Fourier Transform along the
        last axis of `x`.

    Example:

    >>> x = keras.ops.convert_to_tensor([0.0, 1.0, 2.0, 3.0, 4.0])
    >>> istft(stft(x, 1, 1, 1), 1, 1, 1)
    array([0.0, 1.0, 2.0, 3.0, 4.0])
    """
    if not isinstance(sequence_stride, int) or sequence_stride <= 0:
        raise ValueError(
            "`sequence_stride` must be a positive integer. "
            f"Received: sequence_stride={sequence_stride}"
        )
    if any_symbolic_tensors(x):
        return ISTFT(
            sequence_length=sequence_length,
            sequence_stride=sequence_stride,
            fft_length=fft_length,
            length=length,
            window=window,
            center=center,
        ).symbolic_call(x)
    return backend.math.istft(
        x,
        sequence_length=sequence_length,
        sequence_stride=sequence_stride,
        fft_length=fft_length,
        length=length,
        window=window,
        center=center,
    )


class Rsqrt(Operation):
    def call(self, x):
        x = backend.convert_to_tensor(x)
        return backend.math.rsqrt(x)

    def compute_output_spec(self, x):
        return KerasTensor(x.shape, dtype=x.dtype)


@keras_export("keras.ops.rsqrt")
def rsqrt(x):
    """Computes reciprocal of square root of x element-wise.

    Args:
        x: input tensor

    Returns:
        A tensor with the same dtype as `x`.

    Example:

    >>> x = keras.ops.convert_to_tensor([1.0, 10.0, 100.0])
    >>> keras.ops.rsqrt(x)
    array([1.0, 0.31622776, 0.1], dtype=float32)
    """
    if any_symbolic_tensors((x,)):
        return Rsqrt().symbolic_call(x)
    x = backend.convert_to_tensor(x)
    return backend.math.rsqrt(x)


class Erf(Operation):
    def compute_output_spec(self, x):
        return KerasTensor(shape=x.shape, dtype=x.dtype)

    def call(self, x):
        return backend.math.erf(x)


@keras_export("keras.ops.erf")
def erf(x):
    """Computes the error function of `x`, element-wise.

    Args:
        x: Input tensor.

    Returns:
        A tensor with the same dtype as `x`.

    Example:

    >>> x = np.array([-3.0, -2.0, -1.0, 0.0, 1.0])
    >>> keras.ops.erf(x)
    array([-0.99998 , -0.99532, -0.842701,  0.,  0.842701], dtype=float32)
    """
    if any_symbolic_tensors((x,)):
        return Erf().symbolic_call(x)
    x = backend.convert_to_tensor(x)
    return backend.math.erf(x)


class Erfinv(Operation):
    def compute_output_spec(self, x):
        return KerasTensor(shape=x.shape, dtype=x.dtype)

    def call(self, x):
        return backend.math.erfinv(x)


@keras_export("keras.ops.erfinv")
def erfinv(x):
    """Computes the inverse error function of `x`, element-wise.

    Args:
        x: Input tensor.

    Returns:
        A tensor with the same dtype as `x`.

    Example:

    >>> x = np.array([-0.5, -0.2, -0.1, 0.0, 0.3])
    >>> keras.ops.erfinv(x)
    array([-0.47694, -0.17914, -0.08886,  0. ,  0.27246], dtype=float32)
    """
    if any_symbolic_tensors((x,)):
        return Erfinv().symbolic_call(x)
    x = backend.convert_to_tensor(x)
    return backend.math.erfinv(x)


class Logdet(Operation):
    def call(self, x):
        return backend.math.logdet(x)

    def compute_output_spec(self, x):
        return KerasTensor(x.shape[:-2], dtype=x.dtype)


@keras_export(["keras.ops.logdet"])
def logdet(x):
    """Computes log of the determinant of a hermitian positive definite matrix.

    Args:
        x: Input matrix. It must 2D and square.

    Returns:
        The natural log of the determinant of matrix.
    """
    if any_symbolic_tensors((x,)):
        return Logdet().symbolic_call(x)
    return backend.math.logdet(x)


class ViewAsComplex(Operation):
    def call(self, x):
        x = backend.convert_to_tensor(x)
        if len(x.shape) < 1 or x.shape[-1] != 2:
            raise ValueError(
                "Input tensor's last dimension must be 2 (real and imaginary)."
            )
        return x[..., 0] + 1j * x[..., 1]

    def compute_output_spec(self, x):
        return KerasTensor(shape=x.shape[:-1], dtype="complex64")


class ViewAsReal(Operation):
    def call(self, x):
        x = backend.convert_to_tensor(x)
        real_part = backend.numpy.real(x)
        imag_part = backend.numpy.imag(x)
        return backend.numpy.stack((real_part, imag_part), axis=-1)

    def compute_output_spec(self, x):
        return KerasTensor(shape=x.shape + (2,), dtype="float32")


@keras_export("keras.ops.view_as_complex")
def view_as_complex(x):
    """Converts a real tensor with shape `(..., 2)` to a complex tensor,
    where the last dimension represents the real and imaginary components
    of a complex tensor.

    Args:
        x: A real tensor with last dimension of size 2.

    Returns:
        A complex tensor with shape `x.shape[:-1]`.

    Example:

    ```
    >>> import numpy as np
    >>> from keras import ops

    >>> real_imag = np.array([[1.0, 2.0], [3.0, 4.0]])
    >>> complex_tensor = ops.view_as_complex(real_imag)
    >>> complex_tensor
    array([1.+2.j, 3.+4.j])
    ```
    """
    if any_symbolic_tensors((x,)):
        return ViewAsComplex().symbolic_call(x)

    x = backend.convert_to_tensor(x)
    if len(x.shape) < 1 or x.shape[-1] != 2:
        raise ValueError(
            "Last dimension of input must be size 2 (real and imaginary). "
            f"Received shape: {x.shape}"
        )
    real_part = x[..., 0]
    imag_part = x[..., 1]

    return backend.cast(real_part, dtype="complex64") + 1j * backend.cast(
        imag_part, dtype="complex64"
    )


@keras_export("keras.ops.view_as_real")
def view_as_real(x):
    """Converts a complex tensor to a real tensor with shape `(..., 2)`,
    where the last dimension represents the real and imaginary components.

    Args:
        x: A complex tensor.

    Returns:
        A real tensor where the last dimension contains the
        real and imaginary parts.

    Example:
    ```
    >>> import numpy as np
    >>> from keras import ops

    >>> complex_tensor = np.array([1 + 2j, 3 + 4j])
    >>> real = ops.view_as_real(complex_tensor)
    >>> real
    array([[1., 2.],
           [3., 4.]])
    ```
    """
    if any_symbolic_tensors((x,)):
        return ViewAsReal().symbolic_call(x)

    x = backend.convert_to_tensor(x)
    real_part = backend.numpy.real(x)
    imag_part = backend.numpy.imag(x)
    return backend.numpy.stack((real_part, imag_part), axis=-1)
